import os
import posixpath
from collections import defaultdict
from itertools import zip_longest
from typing import DefaultDict, Iterable, List

from attrs import asdict
from fsspec.asyn import get_loop, sync
from sqlalchemy.sql import func
from tqdm import tqdm

from dql.catalog.datasource import DataSource
from dql.client import Client
from dql.data_storage import AbstractDataStorage
from dql.node import DirType, Node, NodeWithPath
from dql.storage import Storage
from dql.utils import suffix_to_number


def check_checksums(nodes):
    for node in nodes:
        if node.n.name and not node.n.checksum:
            raise ValueError(f"Instantiation Error: Missing checksum for node: {node}")


class Listing:
    def __init__(
        self,
        storage: Storage,
        data_storage: AbstractDataStorage,
        client: Client,
    ):
        self.storage = storage
        self.data_storage = data_storage
        self.client = client

    def clone(self) -> "Listing":
        return Listing(
            self.storage, self.data_storage.clone(self.client.uri), self.client
        )

    @property
    def id(self):
        return self.storage.id

    def fetch(self, start_prefix="", partial_id=0):
        sync(get_loop(), self.client.fetch, self, start_prefix, partial_id)

    @staticmethod
    async def _insert_dir(data_storage, parent_id, name, time, path, partial_id):
        return await data_storage.insert_entry(
            {
                "vtype": "",
                "is_dir": True,
                "parent_id": parent_id,
                "path": path,
                "name": name,
                "last_modified": time,
                "checksum": "",
                "etag": "",
                "version": "",
                "is_latest": True,
                "size": 0,
                "owner_name": "",
                "owner_id": "",
                "partial_id": partial_id,
            },
        )

    async def insert_dir(
        self, parent_id, name, time, path, partial_id, data_storage=None
    ):
        return await Listing._insert_dir(
            data_storage or self.data_storage,
            parent_id,
            name,
            time,
            path,
            partial_id,
        )

    async def insert_file(self, parent_id, name, time, path_str, partial_id):
        node = Node(
            0,
            "",
            DirType.FILE,
            parent_id,
            name,
            last_modified=time,
            path_str=path_str,
        )
        await self.insert_file_from_node(node, partial_id)

    async def insert_file_from_node(self, node, partial_id):
        await self.data_storage.insert_entry(
            {**asdict(node), "is_dir": False, "partial_id": partial_id}
        )

    async def insert_root(self, data_storage=None) -> int:
        return await (data_storage or self.data_storage).insert_root()

    def expand_path(self, path) -> List[Node]:
        return self.data_storage.expand_path(path)

    def resolve_path(self, path) -> Node:
        return self.data_storage.get_node_by_path(path)

    def ls_path(self, node, fields):
        if node.vtype == "tar" or node.dir_type == DirType.TAR_ARCHIVE:
            return self.data_storage.select_node_fields_by_parent_path(
                node.path_str, fields
            )
        return self.data_storage.select_node_fields_by_parent_id(node.id, fields)

    def collect_nodes_to_instantiate(
        self,
        sources: Iterable[DataSource],
        recursive=False,
        copy_dir_contents=False,
        relative_path=None,
        from_edql=False,
    ):
        rel_path_elements = relative_path.split("/") if relative_path else []
        all_nodes: List[NodeWithPath] = []
        for src in sources:
            node = src.node
            if recursive and src.is_container():
                dir_path = []
                if not copy_dir_contents:
                    dir_path.append(node.basename)
                subtree_nodes = src.find(sort=["parent_id", "path_str"])
                all_nodes.extend(
                    NodeWithPath(n.n, path=dir_path + n.path) for n in subtree_nodes
                )
            else:
                node_path = []
                if from_edql:
                    for rpe, npe in zip_longest(
                        rel_path_elements, node.path_str.split("/")
                    ):
                        if rpe == npe:
                            continue
                        if npe:
                            node_path.append(npe)
                else:
                    node_path = [node.basename]
                all_nodes.append(NodeWithPath(node, path=node_path))
        return all_nodes

    def instantiate_nodes(
        self,
        all_nodes,
        output,
        cache,
        total_files=None,
        force=False,
        shared_progress_bar=None,
    ):
        check_checksums(all_nodes)
        progress_bar = shared_progress_bar or tqdm(
            desc=f"Instantiating '{output}'",
            unit=" files",
            unit_scale=True,
            unit_divisor=1000,
            total=total_files,
        )

        counter = 0
        for node in all_nodes:
            dst = os.path.join(output, *node.path)
            dst_dir = os.path.dirname(dst)
            os.makedirs(dst_dir, exist_ok=True)
            self.client.instantiate_node(node.n, cache, dst, progress_bar, force)
            counter += 1
            if counter > 1000:
                progress_bar.update(counter)
                counter = 0

        progress_bar.update(counter)

    def download_nodes(
        self,
        sources: Iterable[DataSource],
        cache,
        total_size=None,
        recursive=False,
        shared_progress_bar=None,
    ):
        nodes_by_path: DefaultDict[str, List] = defaultdict(list)
        for src in sources:
            node_path = posixpath.dirname(src.node.path_str)
            if recursive and src.is_container():
                nodes_by_path[node_path].extend(src.find(sort="size desc"))
            else:
                node = src.node
                assert node.name  # for mypy
                nodes_by_path[node_path].append(NodeWithPath(node, path=[node.name]))

        updated_node_lookup = {}
        for group_path, all_nodes in nodes_by_path.items():
            updated_nodes = self.client.fetch_nodes(
                group_path,
                (node.n for node in all_nodes),
                cache,
                self.data_storage,
                total_size,
                shared_progress_bar=shared_progress_bar,
            )
            # This assert is hopefully not necessary, but is here mostly to
            # ensure this functionality stays consistent (and catch any bugs)
            assert len(updated_nodes) <= len(all_nodes)

            for node in updated_nodes:
                updated_node_lookup[node.id] = node

        return {i: n.checksum for i, n in updated_node_lookup.items()}

    def find(
        self,
        node,
        fields,
        names=None,
        inames=None,
        paths=None,
        ipaths=None,
        size=None,
        type=None,  # pylint: disable=redefined-builtin
        jmespath=None,
        order_by=None,
    ):
        n = self.data_storage.nodes
        conds = []
        if names:
            for name in names:
                conds.append(n.c.name.op("GLOB")(name))
        if inames:
            for iname in inames:
                conds.append(func.lower(n.c.name).op("GLOB")(iname.lower()))
        if paths:
            for path in paths:
                conds.append(n.c.path_str.op("GLOB")(path))
        if ipaths:
            for ipath in ipaths:
                conds.append(func.lower(n.c.path_str).op("GLOB")(ipath.lower()))

        if size is not None:
            size_limit = suffix_to_number(size)
            if size_limit >= 0:
                conds.append(n.c.size >= size_limit)
            else:
                conds.append(n.c.size <= -size_limit)

        return self.data_storage.find(
            node,
            fields,
            jmespath=jmespath,
            type=type,
            conds=conds,
            order_by=order_by,
        )

    def du(self, node: Node, count_files: bool = False):
        return self.data_storage.size(node, count_files)



import hashlib
import sys
from functools import partial

if sys.version_info < (3, 9):
    md5 = hashlib.md5
else:
    md5 = partial(hashlib.md5, usedforsecurity=False)

BUFSIZE = 2**18


def file_digest(fileobj):
    """Calculate the digest of a file-like object."""
    buf = bytearray(BUFSIZE)  # Reusable buffer to reduce allocations.
    view = memoryview(buf)
    digestobj = md5()
    # From 3.11's hashlib.filedigest()
    while True:
        size = fileobj.readinto(buf)
        if size == 0:
            break  # EOF
        digestobj.update(view[:size])
    return digestobj.hexdigest()



from typing import cast

from s3fs import S3FileSystem

from .fsspec import DELIMITER, TIME_ZERO, FSSpecClient

UPDATE_CHUNKSIZE = 1000


class ClientS3(FSSpecClient):
    FS_CLASS = S3FileSystem
    PREFIX = "s3://"
    protocol = "s3"

    @classmethod
    def create_fs(cls, **kwargs) -> S3FileSystem:
        if "aws_anon" in kwargs:
            kwargs.setdefault("anon", kwargs.pop("aws_anon"))
        if "aws_endpoint_url" in kwargs:
            kwargs.setdefault("client_kwargs", {}).setdefault(
                "endpoint_url", kwargs.pop("aws_endpoint_url")
            )
        if "aws_key" in kwargs:
            kwargs.setdefault("key", kwargs.pop("aws_key"))
        if "aws_secret" in kwargs:
            kwargs.setdefault("secret", kwargs.pop("aws_secret"))
        if "aws_token" in kwargs:
            kwargs.setdefault("token", kwargs.pop("aws_token"))

        # caching bucket regions to use the right one in signed urls, otherwise
        # it tries to randomly guess and creates wrong signature
        kwargs.setdefault("cache_regions", True)

        # We want to use newer v4 signature version since regions added after
        # 2014 are not going to support v2 which is the older one.
        # All regions support v4.
        kwargs.setdefault("config_kwargs", {}).setdefault("signature_version", "s3v4")

        return cast(S3FileSystem, super().create_fs(**kwargs))

    def url(self, path: str, expires: int = 3600) -> str:
        """
        Returns a signed URL of specific file in a bucket that lasts for some time,
        which is defined by expires argument.
        """
        return self.fs.url(self.get_full_path(path), expires=expires)

    async def _fetch_dir(self, dir_id, prefix, pbar, listing, data_storage, partial_id):
        if prefix:
            prefix = prefix.lstrip(DELIMITER) + DELIMITER
        files = []
        subdirs = set()
        subdir_names = set()
        ignore_count = 0
        pbar_updated = 0
        # pylint:disable-next=protected-access
        async for info in self.fs._iterdir(self.name, prefix=prefix, versions=True):
            full_path = info["name"]
            _, subprefix, _ = self.fs.split_path(info["name"])
            if info["type"] == "directory":
                name = full_path.split(DELIMITER)[-1]
                new_dir_id = await listing.insert_dir(
                    dir_id,
                    name,
                    TIME_ZERO,
                    subprefix,
                    partial_id,
                    data_storage=data_storage,
                )
                subdirs.add((new_dir_id, subprefix))
                subdir_names.add(name)
            else:
                files.append(self._dict_from_info(info, dir_id, subprefix, partial_id))
                pbar_update_count = len(files) - pbar_updated
                if pbar_update_count >= UPDATE_CHUNKSIZE:
                    pbar.update(pbar_update_count)
                    pbar_updated += pbar_update_count
        for f in files:
            if not f["name"] or f["name"] in subdir_names:
                # Set files that conflict with directories as not valid
                # Files without a name are of a prefix like "dir/" where the name
                # ends up being empty after the split on "/"
                # And the conflicting files have prefixes like "dir" where there is
                # also an "dir" directory, for example "dir/file1" as well.
                f["valid"] = False
                ignore_count += 1
        if files:
            await data_storage.insert_entries(files)
            await data_storage.update_last_inserted_at()
            pbar_update_count = len(files) - pbar_updated
            pbar.update(pbar_update_count)
        pbar.update(len(subdirs))
        return subdirs, ignore_count

    @staticmethod
    def clean_s3_version(ver):
        return ver if ver != "null" else ""

    def _dict_from_info(self, v, parent_id, path, partial_id):
        return {
            "is_dir": False,
            "parent_id": parent_id,
            "path": path,
            "name": v.get("Key", "").split(DELIMITER)[-1],
            # 'expires': expires,
            "checksum": "",
            "etag": v.get("ETag", "").strip('"'),
            "version": ClientS3.clean_s3_version(v.get("VersionId", "")),
            "is_latest": v.get("IsLatest", True),
            "last_modified": v.get("LastModified", ""),
            "size": v.get("Size", ""),
            # 'storage_class': v.get('StorageClass'),
            "owner_name": v.get("Owner", {}).get("DisplayName", ""),
            "owner_id": v.get("Owner", {}).get("ID", ""),
            "anno": None,
            "partial_id": partial_id,
        }



import os
from subprocess import check_output

import pytest
from dulwich.porcelain import clone
from packaging import version
from pytest_virtualenv import VirtualEnv


@pytest.fixture
def bucket():
    return "s3://noaa-bathymetry-pds/"


def pytest_generate_tests(metafunc):
    str_revs = metafunc.config.getoption("--dql-revs")
    revs = str_revs.split(",") if str_revs else [None]
    if "dql_rev" in metafunc.fixturenames:
        metafunc.parametrize("dql_rev", revs, scope="session")


@pytest.fixture(scope="session", name="make_dql_venv")
def fixture_make_dql_venv(tmp_path_factory):
    def _make_dql_venv(name):
        venv_dir = tmp_path_factory.mktemp(f"dql-venv-{name}")
        return VirtualEnv(workspace=venv_dir)

    return _make_dql_venv


@pytest.fixture(scope="session", name="dql_venvs")
def fixture_dql_venvs():
    return {}


@pytest.fixture(scope="session", name="dql_git_repo")
def fixture_dql_git_repo(tmp_path_factory, test_config):
    url = test_config.dql_git_repo

    if os.path.isdir(url):
        return url

    tmp_path = os.fspath(tmp_path_factory.mktemp("dql-git-repo"))
    clone(url, tmp_path)

    return tmp_path


@pytest.fixture(scope="session", name="dql_bin")
def fixture_dql_bin(
    dql_rev,
    dql_venvs,
    make_dql_venv,
    dql_git_repo,
    test_config,
):
    if dql_rev:
        venv = dql_venvs.get(dql_rev)
        if not venv:
            venv = make_dql_venv(dql_rev)
            venv.run("pip install -U pip")
            venv.run(f"pip install git+file://{dql_git_repo}@{dql_rev}")
            dql_venvs[dql_rev] = venv
        dql_bin = venv.virtualenv / "bin" / "dql"
    else:
        dql_bin = test_config.dql_bin

    def _dql_bin(*args):
        return check_output([dql_bin, *args], text=True)

    actual = version.parse(_dql_bin("--version"))
    _dql_bin.version = (actual.major, actual.minor, actual.micro)

    return _dql_bin


@pytest.fixture(scope="function", name="make_bench")
def fixture_make_bench(request):
    def _make_bench(name):
        import pytest_benchmark.plugin

        # hack from https://github.com/ionelmc/pytest-benchmark/issues/166
        bench = pytest_benchmark.plugin.benchmark.__pytest_wrapped__.obj(request)

        suffix = f"-{name}"

        def add_suffix(_name):
            start, sep, end = _name.partition("[")
            return start + suffix + sep + end

        bench.name = add_suffix(bench.name)
        bench.fullname = add_suffix(bench.fullname)

        return bench

    return _make_bench


@pytest.fixture(
    scope="function", params=[pytest.param(None, marks=pytest.mark.benchmark)]
)
def bench_dql(dql_bin, make_bench):
    def _bench_dql(*args, **kwargs):
        name = kwargs.pop("name", None)
        name = f"-{name}" if name else ""
        bench = make_bench(args[0] + name)
        return bench.pedantic(dql_bin, args=args, **kwargs)

    return _bench_dql



import asyncio
import io
import tarfile
from tarfile import DIRTYPE, TarInfo


def insert_entries(data_storage, entries):
    """
    Takes a list of entries and inserts them synchronously.

    This assumes that, for every entry in entries, all of its parent
    directories also have an entry.
    """
    entries_by_path = [(tuple(e["path"].split("/")), e) for e in entries]
    entries_by_path.sort(key=lambda e: (len(e[0]), not e[1]["is_dir"]))

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        dir_ids = {(): asyncio.run(data_storage.insert_root())}
        for path_parts, entry in entries_by_path:
            entry = {
                **entry,
                "parent_id": dir_ids[path_parts[:-1]],
                "partial_id": 0,
            }
            node_id = loop.run_until_complete(data_storage.insert_entry(entry))
            if entry["is_dir"]:
                dir_ids[path_parts] = node_id
    finally:
        asyncio.set_event_loop(None)
        loop.close()


DEFAULT_TREE = {
    "description": "Cats and Dogs",
    "cats": {
        "cat1": "meow",
        "cat2": "mrow",
    },
    "dogs": {
        "dog1": "woof",
        "dog2": "arf",
        "dog3": "bark",
        "others": {"dog4": "ruff"},
    },
}


def instantiate_tree(path, tree):
    for key, value in tree.items():
        if isinstance(value, str):
            (path / key).write_text(value)
        elif isinstance(value, bytes):
            (path / key).write_bytes(value)
        elif isinstance(value, dict):
            (path / key).mkdir()
            instantiate_tree(path / key, value)
        else:
            raise TypeError(f"{value=}")


def tree_from_path(path, binary=False):
    tree = {}
    for child in path.iterdir():
        if child.is_dir():
            tree[child.name] = tree_from_path(child, binary)
        else:
            if binary:
                tree[child.name] = child.read_bytes()
            else:
                tree[child.name] = child.read_text()
    return tree


def uppercase_scheme(uri: str) -> str:
    """
    Makes scheme (or protocol) of an url uppercased
    e.g s3://bucket_name -> S3://bucket_name
    """
    return f'{uri.split(":")[0].upper()}:{":".join(uri.split(":")[1:])}'


def make_tar(tree):
    with io.BytesIO() as tmp:
        with tarfile.open(fileobj=tmp, mode="w") as archive:
            write_tar(tree, archive)
        return tmp.getvalue()


def write_tar(tree, archive, curr_dir=""):
    for key, value in tree.items():
        name = f"{curr_dir}/{key}" if curr_dir else key
        if isinstance(value, str):
            value = value.encode("utf-8")
        if isinstance(value, bytes):
            info = TarInfo(name)
            info.size = len(value)
            f = io.BytesIO(value)
            archive.addfile(info, f)
        elif isinstance(value, dict):
            info = TarInfo(name)
            info.type = DIRTYPE
            archive.addfile(info, io.BytesIO())
            write_tar(value, archive, name)


TARRED_TREE = {"animals.tar": make_tar(DEFAULT_TREE)}



import os
from importlib import import_module

from dql.catalog import Catalog
from dql.data_storage import SQLiteDataStorage
from dql.utils import get_envs_by_prefix

ADAPTER_ARG_PREFIX = "DQL_ADAPTER_ARG_"


def get_catalog(client_config=None) -> Catalog:
    """
    Function that creates Catalog instance with appropriate data storage class
    Data storage class can be provided with env variable DQL_DB_ADAPTER and if
    not provided, default one is used

    If data storage class expects some kwargs, they can be provided via
    env variables by adding them with prefix DQL_DB_ADAPTER_ARG_ and
    name of variable after, e.g if it accepts team_id as kwargs we can provide
    DQL_DB_ADAPTER_ARG_TEAM_ID=12345 env variable
    """
    db_adapter_import_path = os.environ.get("DQL_DB_ADAPTER")
    adapter_arg_envs = get_envs_by_prefix(ADAPTER_ARG_PREFIX)
    # convert env variable names to keyword argument names by lowercasing it
    adapter_args = {k.lower(): v for k, v in adapter_arg_envs.items()}

    if db_adapter_import_path:
        # DB Adapter paths are specified as (for example):
        # dql.data_storage.SQLiteDataStorage
        if "." not in db_adapter_import_path:
            raise RuntimeError(
                f"Invalid DQL_DB_ADAPTER import path: {db_adapter_import_path}"
            )
        module_name, _, class_name = db_adapter_import_path.rpartition(".")
        db_adapter = import_module(module_name)
        db_adapter_class = getattr(db_adapter, class_name)
    else:
        db_adapter_class = SQLiteDataStorage
    return Catalog(db_adapter_class(**adapter_args), client_config=client_config)



import jmespath
from jmespath.visitor import Options, TreeInterpreter

from dql.jmes_sql.field_ops import FieldOps
from dql.jmes_sql.functions_sql import FunctionsSql


class Transpiler(TreeInterpreter):
    COMPARATOR_FUNC = {
        "eq": FieldOps.equals,
        "ne": FieldOps.not_equals,
        "lt": FieldOps.op_lt,
        "gt": FieldOps.op_gt,
        "lte": FieldOps.op_le,
        "gte": FieldOps.op_ge,
    }

    def __init__(self, table, json_column, all_columns):
        self.original_table = table
        self.json_column = json_column
        self.all_columns = all_columns

        super().__init__(options=Options(custom_functions=FunctionsSql(self)))

        self._views = []
        self._views_counter = 0
        self._last_view_name = json_column

    @staticmethod
    def execute_query(con, query, table, json_column, all_columns):
        sql = Transpiler.build_sql(query, table, json_column, all_columns)
        return con.executescript(sql)

    @staticmethod
    def build_sql(query, table, json_column, all_columns):
        ts = Transpiler(table, json_column, all_columns)
        sql, view_name = ts.translate(query)
        q = f"""
            {sql}

            SELECT *
            FROM {view_name}
        """
        return q

    def default_visit(self, node, *args, **kwargs):
        raise NotImplementedError("default_visit")

    @property
    def all_columns_in_sql(self):
        return ", ".join(self.all_columns)

    def _append_condition_view(self, condition):
        new_view_name = f"V{self._views_counter}"
        self._views.append(
            f"""
            CREATE TEMP VIEW {new_view_name} AS
            SELECT * FROM '{self._last_view_name}'
            {'WHERE ' + condition if condition else ''};
        """
        )

        self._last_view_name = new_view_name
        self._views_counter += 1

    def _append_initial_view(self):
        new_view_name = f"V{self._views_counter}"
        self._views.append(
            f"""
            CREATE TEMP VIEW {new_view_name} AS
            SELECT {self.all_columns_in_sql},
                   {self.json_column} AS {FieldOps.CUST_FIELD}
            FROM '{self.original_table}'
        """
        )

        self._views.append("""DROP MACRO IF EXISTS PROJECT""")
        self._views.append(
            """
            CREATE MACRO PROJECT(arr, field) AS
                list_transform(from_json(arr, '["json"]'),
                               x -> json_extract_string(json(x), field))
        """
        )

        self._last_view_name = new_view_name
        self._views_counter += 1

    def _append_cust_field_view(self, cust_field):
        new_view_name = f"V{self._views_counter}"
        self._views.append(
            f"""
            CREATE TEMP VIEW {new_view_name} AS
            SELECT {self.all_columns_in_sql},
                   {cust_field} AS {FieldOps.CUST_FIELD}
            FROM '{self._last_view_name}'
        """
        )

        self._last_view_name = new_view_name
        self._views_counter += 1

    def _append_final_view(self):
        new_view_name = f"V{self._views_counter}"
        self._views.append(
            f"""
            CREATE TEMP VIEW {new_view_name} AS
            SELECT {self.all_columns_in_sql}
            FROM '{self._last_view_name}'
            WHERE {FieldOps.CUST_FIELD}
        """
        )

        self._last_view_name = new_view_name
        self._views_counter += 1

    def translate(self, query):
        self._append_initial_view()

        tree = jmespath.compile(query).parsed
        result = self.visit(tree, "")

        self._append_cust_field_view(result)
        self._append_final_view()

        prepare_views = ";".join(self._views) + ";"

        return prepare_views, self._last_view_name

    def visit_field(self, node, value):
        try:
            node_value = node["value"]

            field = FieldOps.get_field_name(value)
            if field:
                return f"{value}->>'{node_value}'"

            return FieldOps.str_to_filed(node_value)
        except AttributeError:
            return None

    def visit_function_expression(self, node, value):
        resolved_args = []
        for child in node["children"]:
            current = self.visit(child, value)
            resolved_args.append(current)
        return self._functions.call_function(node["value"], resolved_args)

    def visit_index(self, node, value):
        return value + "->>" + str(node["value"])

    def visit_projection(self, node, value):
        base = self.visit(node["children"][0], value)
        field = FieldOps.resolve_filed(base)
        if not field:
            return TreeInterpreter.visit_projection(self, node, value)

        curr = self.visit(node["children"][1], "")
        curr_field_name = FieldOps.get_field_name(curr)

        return f"{FieldOps.CMD_PREFIX}PROJECT({field}, {curr_field_name})"



import io
import os
import uuid
from unittest.mock import patch

import pytest
import sqlalchemy as sa

from dql.data_storage.sqlite import SQLiteDataStorage
from dql.dataset import Status as DatasetStatus
from dql.error import DatasetNotFoundError
from dql.loader import DataView

# pylint: disable=redefined-outer-name,unused-argument,protected-access


def skip_if_not_sqlite():
    if os.environ.get("DQL_DB_ADAPTER"):
        pytest.skip("This test is not supported on other data storages")


@pytest.fixture
def empty_shadow_dataset(listed_bucket, cloud_test_catalog):
    shadow_dataset_name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog

    catalog.create_shadow_dataset(
        shadow_dataset_name,
        [],
        client_config=cloud_test_catalog.client_config,
        populate=False,
    )

    return catalog.data_storage.get_dataset(shadow_dataset_name)


@pytest.fixture
def cats_shadow_dataset(listed_bucket, cloud_test_catalog):
    shadow_dataset_name = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    catalog.create_shadow_dataset(
        shadow_dataset_name,
        [f"{src_uri}/cats/*"],
        client_config=cloud_test_catalog.client_config,
        recursive=True,
    )

    return catalog.data_storage.get_dataset(shadow_dataset_name)


@pytest.fixture
def dogs_registered_dataset(cloud_test_catalog, dogs_shadow_dataset):
    catalog = cloud_test_catalog.catalog
    catalog.register_shadow_dataset(
        dogs_shadow_dataset.name,
        description="dogs dataset",
        labels=["dogs", "dataset"],
    )

    return catalog.data_storage.get_dataset(dogs_shadow_dataset.name)


@pytest.fixture
def cats_registered_dataset(cloud_test_catalog, cats_shadow_dataset):
    catalog = cloud_test_catalog.catalog
    catalog.register_shadow_dataset(
        cats_shadow_dataset.name,
        description="cats dataset",
        labels=["cats", "dataset"],
    )

    return catalog.data_storage.get_dataset(cats_shadow_dataset.name)


@pytest.fixture
def mock_insert_into_shadow_dataset():
    with patch.object(
        SQLiteDataStorage,
        "insert_into_shadow_dataset",
        side_effect=RuntimeError("Error"),
    ) as m:
        yield m


def test_get_dataset(cloud_test_catalog, dogs_shadow_dataset):
    catalog = cloud_test_catalog.catalog

    dataset = catalog.get_dataset(dogs_shadow_dataset.name)
    assert dataset.name == dogs_shadow_dataset.name

    with pytest.raises(DatasetNotFoundError):
        catalog.get_dataset("wrong name")


# Returns None if the table does not exist
def get_table_row_count(data_storage, table_name):
    if not data_storage.has_table(table_name):
        return None
    query = sa.select(sa.func.count()).select_from(sa.text(table_name))
    return next(data_storage.execute(query), (None,))[0]


def test_creating_shadow_dataset(listed_bucket, cloud_test_catalog):
    shadow_dataset_name = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    dataset = catalog.create_shadow_dataset(
        shadow_dataset_name,
        [f"{src_uri}/dogs/*"],
        client_config=cloud_test_catalog.client_config,
        recursive=True,
    )

    assert dataset.name == shadow_dataset_name
    assert dataset.description is None
    assert dataset.versions is None
    assert dataset.labels == []
    assert dataset.shadow is True
    assert dataset.status == DatasetStatus.COMPLETE
    assert dataset.created_at
    assert dataset.finished_at

    dataset_table_name = catalog.data_storage.dataset_table_name(dataset.id)
    assert get_table_row_count(catalog.data_storage, dataset_table_name)


def test_creating_shadow_dataset_failed(
    listed_bucket, cloud_test_catalog, mock_insert_into_shadow_dataset
):
    shadow_dataset_name = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog
    with pytest.raises(RuntimeError):
        catalog.create_shadow_dataset(
            shadow_dataset_name,
            [f"{src_uri}/dogs/*"],
            client_config=cloud_test_catalog.client_config,
            recursive=True,
        )
    dataset = catalog.data_storage.get_dataset(shadow_dataset_name)

    assert dataset.name == shadow_dataset_name
    assert dataset.status == DatasetStatus.FAILED
    assert dataset.created_at
    assert dataset.finished_at

    dataset_table_name = catalog.data_storage.dataset_table_name(dataset.id)
    assert get_table_row_count(catalog.data_storage, dataset_table_name) == 0


def test_creating_empty_dataset(listed_bucket, cloud_test_catalog):
    shadow_dataset_name = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    dataset = catalog.create_shadow_dataset(
        shadow_dataset_name,
        [f"{src_uri}/dogs/*"],
        client_config=cloud_test_catalog.client_config,
        recursive=True,
        populate=False,
    )

    assert dataset.name == shadow_dataset_name
    assert dataset.status == DatasetStatus.CREATED
    assert dataset.shadow is True
    assert dataset.created_at
    assert not dataset.finished_at

    dataset_table_name = catalog.data_storage.dataset_table_name(dataset.id)
    assert get_table_row_count(catalog.data_storage, dataset_table_name) is None


def test_creating_dataset_after_empty(listed_bucket, cloud_test_catalog):
    shadow_dataset_name = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    dataset = catalog.create_shadow_dataset(
        shadow_dataset_name,
        [f"{src_uri}/dogs/*"],
        client_config=cloud_test_catalog.client_config,
        recursive=True,
        populate=False,
    )

    assert dataset.status == DatasetStatus.CREATED

    dataset = catalog.create_shadow_dataset(
        shadow_dataset_name,
        [f"{src_uri}/dogs/*"],
        client_config=cloud_test_catalog.client_config,
        recursive=True,
    )

    assert dataset.status == DatasetStatus.COMPLETE
    assert dataset.created_at
    assert dataset.finished_at

    dataset_table_name = catalog.data_storage.dataset_table_name(dataset.id)
    assert get_table_row_count(catalog.data_storage, dataset_table_name)


def test_creating_shadow_dataset_whole_bucket(listed_bucket, cloud_test_catalog):
    shadow_dataset_name_1 = uuid.uuid4().hex
    shadow_dataset_name_2 = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    catalog.create_shadow_dataset(
        shadow_dataset_name_1,
        [f"{src_uri}"],
        client_config=cloud_test_catalog.client_config,
        recursive=True,
    )
    catalog.create_shadow_dataset(
        shadow_dataset_name_2,
        [f"{src_uri}/"],
        client_config=cloud_test_catalog.client_config,
        recursive=True,
    )

    expected_rows = {
        "description",
        "cat1",
        "cat2",
        "dog1",
        "dog2",
        "dog3",
        "dog4",
    }

    assert {
        r.name for r in catalog.ls_dataset_rows(shadow_dataset_name_1)
    } == expected_rows

    assert {
        r.name for r in catalog.ls_dataset_rows(shadow_dataset_name_2)
    } == expected_rows


def test_registering_dataset(cloud_test_catalog, dogs_shadow_dataset):
    catalog = cloud_test_catalog.catalog

    catalog.register_shadow_dataset(
        dogs_shadow_dataset.name,
        description="dogs dataset",
        labels=["dogs", "dataset"],
    )

    dataset = catalog.data_storage.get_dataset(dogs_shadow_dataset.name)
    dataset_table_name = catalog.data_storage.dataset_table_name(dataset.id, 1)
    assert dataset.name == dogs_shadow_dataset.name
    assert dataset.description == "dogs dataset"
    assert dataset.versions_values == [1]
    assert dataset.labels == ["dogs", "dataset"]
    assert dataset.shadow is False
    assert dataset.status == DatasetStatus.COMPLETE
    assert get_table_row_count(catalog.data_storage, dataset_table_name)


def test_registering_dataset_with_new_name(cloud_test_catalog, dogs_shadow_dataset):
    new_dataset_name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog

    catalog.register_shadow_dataset(
        dogs_shadow_dataset.name,
        registered_name=new_dataset_name,
        description="dogs dataset",
        labels=["dogs", "dataset"],
    )
    dataset = catalog.data_storage.get_dataset(new_dataset_name)
    assert dataset
    dataset_table_name = catalog.data_storage.dataset_table_name(dataset.id, 1)
    assert dataset.name == new_dataset_name
    assert get_table_row_count(catalog.data_storage, dataset_table_name)


def test_registering_dataset_with_custom_version(
    cloud_test_catalog, dogs_shadow_dataset
):
    catalog = cloud_test_catalog.catalog

    catalog.register_shadow_dataset(
        dogs_shadow_dataset.name,
        version=5,
        description="dogs dataset",
        labels=["dogs", "dataset"],
    )

    dataset = catalog.data_storage.get_dataset(dogs_shadow_dataset.name)
    assert dataset.versions_values == [5]


def test_registering_dataset_as_version_of_another_registered(
    cloud_test_catalog, dogs_registered_dataset, cats_shadow_dataset
):
    catalog = cloud_test_catalog.catalog

    catalog.register_shadow_dataset(
        cats_shadow_dataset.name,
        registered_name=dogs_registered_dataset.name,
        version=3,
    )

    dogs_dataset = catalog.data_storage.get_dataset(dogs_registered_dataset.name)
    assert dogs_dataset.versions_values == [1, 3]
    # checking newly created dogs version 3 data
    assert {
        r.name for r in catalog.ls_dataset_rows(dogs_registered_dataset.name, version=3)
    } == {
        "cat1",
        "cat2",
    }

    # assert cats shadow dataset is removed
    cats_dataset_name = catalog.data_storage.dataset_table_name(cats_shadow_dataset.id)
    with pytest.raises(DatasetNotFoundError):
        catalog.data_storage.get_dataset(cats_shadow_dataset.name)
    assert get_table_row_count(catalog.data_storage, cats_dataset_name) is None


def test_removing_dataset(cloud_test_catalog, dogs_shadow_dataset):
    catalog = cloud_test_catalog.catalog

    dataset_table_name = catalog.data_storage.dataset_table_name(dogs_shadow_dataset.id)
    assert get_table_row_count(catalog.data_storage, dataset_table_name)

    catalog.remove_dataset(dogs_shadow_dataset.name)
    with pytest.raises(DatasetNotFoundError):
        catalog.data_storage.get_dataset(dogs_shadow_dataset.name)

    assert get_table_row_count(catalog.data_storage, dataset_table_name) is None


def test_edit_dataset(cloud_test_catalog, dogs_registered_dataset):
    dataset_new_name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog

    catalog.edit_dataset(
        dogs_registered_dataset.name,
        new_name=dataset_new_name,
        description="new description",
        labels=["cats", "birds"],
    )

    dataset = catalog.data_storage.get_dataset(dataset_new_name)
    assert dataset.versions_values == [1]
    assert dataset.name == dataset_new_name
    assert dataset.description == "new description"
    assert dataset.labels == ["cats", "birds"]


def test_edit_dataset_remove_labels_and_description(
    cloud_test_catalog, dogs_registered_dataset
):
    dataset_new_name = uuid.uuid4().hex
    catalog = cloud_test_catalog.catalog

    catalog.edit_dataset(
        dogs_registered_dataset.name,
        new_name=dataset_new_name,
        description="",
        labels=[],
    )

    dataset = catalog.data_storage.get_dataset(dataset_new_name)
    assert dataset.versions_values == [1]
    assert dataset.name == dataset_new_name
    assert dataset.description == ""
    assert dataset.labels == []


def test_ls_dataset_rows(cloud_test_catalog, dogs_registered_dataset):
    catalog = cloud_test_catalog.catalog

    assert {
        r.name for r in catalog.ls_dataset_rows(dogs_registered_dataset.name, version=1)
    } == {
        "dog1",
        "dog2",
        "dog3",
        "dog4",
    }


def test_merge_datasets_shadow_to_registered(
    cloud_test_catalog, dogs_registered_dataset, cats_shadow_dataset
):
    catalog = cloud_test_catalog.catalog
    catalog.merge_datasets(
        cats_shadow_dataset.name, dogs_registered_dataset.name, dst_version=2
    )

    dogs_dataset = catalog.data_storage.get_dataset(dogs_registered_dataset.name)
    assert dogs_dataset.versions_values == [1, 2]

    # making sure version 1 is not changed
    assert {
        r.name for r in catalog.ls_dataset_rows(dogs_registered_dataset.name, version=1)
    } == {
        "dog1",
        "dog2",
        "dog3",
        "dog4",
    }

    assert {
        r.name for r in catalog.ls_dataset_rows(dogs_registered_dataset.name, version=2)
    } == {
        "dog1",
        "dog2",
        "dog3",
        "dog4",
        "cat1",
        "cat2",
    }


def test_merge_datasets_registered_to_registered(
    cloud_test_catalog, dogs_registered_dataset, cats_registered_dataset
):
    catalog = cloud_test_catalog.catalog
    catalog.merge_datasets(
        cats_registered_dataset.name,
        dogs_registered_dataset.name,
        src_version=1,
        dst_version=2,
    )

    dogs_dataset = catalog.data_storage.get_dataset(dogs_registered_dataset.name)
    assert dogs_dataset.versions_values == [1, 2]

    assert {
        r.name for r in catalog.ls_dataset_rows(dogs_registered_dataset.name, version=2)
    } == {
        "dog1",
        "dog2",
        "dog3",
        "dog4",
        "cat1",
        "cat2",
    }


def test_merge_datasets_shadow_to_shadow(
    cloud_test_catalog, dogs_shadow_dataset, cats_shadow_dataset
):
    catalog = cloud_test_catalog.catalog
    catalog.merge_datasets(
        cats_shadow_dataset.name,
        dogs_shadow_dataset.name,
    )

    dogs_dataset = catalog.data_storage.get_dataset(dogs_shadow_dataset.name)
    assert dogs_dataset.shadow is True  # dataset stays shadow

    assert {r.name for r in catalog.ls_dataset_rows(dogs_shadow_dataset.name)} == {
        "dog1",
        "dog2",
        "dog3",
        "dog4",
        "cat1",
        "cat2",
    }


def test_merge_datasets_registered_to_shadow(
    cloud_test_catalog, dogs_shadow_dataset, cats_registered_dataset
):
    catalog = cloud_test_catalog.catalog
    catalog.merge_datasets(
        cats_registered_dataset.name,
        dogs_shadow_dataset.name,
        src_version=1,
    )

    dogs_dataset = catalog.data_storage.get_dataset(dogs_shadow_dataset.name)
    assert dogs_dataset.shadow is True  # dataset stays shadow

    assert {r.name for r in catalog.ls_dataset_rows(dogs_shadow_dataset.name)} == {
        "dog1",
        "dog2",
        "dog3",
        "dog4",
        "cat1",
        "cat2",
    }


def test_merge_datasets_shadow_to_empty_shadow_without_rows_table(
    cloud_test_catalog, empty_shadow_dataset, cats_shadow_dataset
):
    catalog = cloud_test_catalog.catalog
    catalog.merge_datasets(
        cats_shadow_dataset.name,
        empty_shadow_dataset.name,
    )

    empty_dataset = catalog.data_storage.get_dataset(empty_shadow_dataset.name)
    assert empty_dataset.shadow is True  # dataset stays shadow

    assert {r.name for r in catalog.ls_dataset_rows(empty_shadow_dataset.name)} == {
        "cat1",
        "cat2",
    }


def test_merge_datasets_empty_shadow_without_table_to_shadow(
    cloud_test_catalog, empty_shadow_dataset, cats_shadow_dataset
):
    catalog = cloud_test_catalog.catalog
    catalog.merge_datasets(
        empty_shadow_dataset.name,
        cats_shadow_dataset.name,
    )

    cat_dataset = catalog.data_storage.get_dataset(cats_shadow_dataset.name)
    assert cat_dataset.shadow is True  # dataset stays shadow

    assert {r.name for r in catalog.ls_dataset_rows(cats_shadow_dataset.name)} == {
        "cat1",
        "cat2",
    }


def to_str(buf) -> str:
    return io.TextIOWrapper(buf, encoding="utf8").read()


def test_loader_from_dataset(cloud_test_catalog, dogs_shadow_dataset):
    ctc = cloud_test_catalog

    def transform(row, sample):
        return sample, row.name[-1]

    ds = DataView.from_dataset(
        dogs_shadow_dataset.name,
        reader=to_str,
        transform=transform,
        catalog=ctc.catalog,
        client_config=ctc.client_config,
    )
    assert set(ds) == {("woof", "1"), ("arf", "2"), ("bark", "3"), ("ruff", "4")}


def test_loader_from_dataset_split(cloud_test_catalog, dogs_shadow_dataset):
    ctc = cloud_test_catalog

    def transform(row, sample):
        return sample, row.name[-1]

    N = 5
    all_data = []
    for i in range(N):
        ds = DataView.from_dataset(
            dogs_shadow_dataset.name,
            reader=to_str,
            transform=transform,
            num_workers=N,
            worker_id=i,
            catalog=ctc.catalog,
            client_config=ctc.client_config,
        )
        all_data.extend(ds)

    assert len(all_data) == 4
    assert set(all_data) == {("woof", "1"), ("arf", "2"), ("bark", "3"), ("ruff", "4")}


def test_dataset_row(cloud_test_catalog, dogs_shadow_dataset):
    skip_if_not_sqlite()
    catalog = cloud_test_catalog.catalog
    row = catalog.dataset_row(dogs_shadow_dataset.name, 1)
    assert row.id == 1


@pytest.mark.parametrize("tree", [{str(i): str(i) for i in range(50)}], indirect=True)
def test_row_random(cloud_test_catalog):
    # Note: this is technically a probabilistic test, but the probability
    # of accidental failure is < 1e-10
    ctc = cloud_test_catalog
    catalog = ctc.catalog
    catalog.index([ctc.src_uri], client_config=ctc.client_config)
    catalog.create_shadow_dataset(
        "test", [ctc.src_uri], client_config=ctc.client_config
    )
    random_values = [row.random for row in catalog.ls_dataset_rows("test")]

    # Random values are unique
    assert len(set(random_values)) == len(random_values)

    RAND_MAX = 2**63
    # Values are drawn uniformly from range(2**63)
    assert 0 <= min(random_values) < 0.4 * RAND_MAX
    assert 0.6 * RAND_MAX < max(random_values) < RAND_MAX

    # Creating a new dataset preserves random values
    catalog.create_shadow_dataset(
        "test2", [ctc.src_uri], client_config=ctc.client_config
    )
    random_values2 = [row.random for row in catalog.ls_dataset_rows("test2")]
    assert random_values2 == random_values


def test_create_shadow_dataset_from_storage(listed_bucket, cloud_test_catalog):
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    catalog.index([src_uri], client_config=cloud_test_catalog.client_config)
    # Add a custom column to the bucket.
    catalog.data_storage.add_bucket_signal_column(src_uri, "test_col", sa.Integer())

    # ensure the altered table's Table object was reloaded properly
    node_columns = [c.name for c in catalog.data_storage.nodes_table(src_uri).columns]
    assert "test_col" in node_columns

    shadow_dataset_name = uuid.uuid4().hex
    dataset = catalog.create_shadow_dataset(
        shadow_dataset_name,
        [f"{src_uri}/dogs/*"],
        client_config=cloud_test_catalog.client_config,
        recursive=True,
    )

    assert dataset.name == shadow_dataset_name
    assert dataset.status == DatasetStatus.COMPLETE


def test_dataset_stats_shadow_ds(cloud_test_catalog, dogs_shadow_dataset):
    catalog = cloud_test_catalog.catalog
    stats = catalog.dataset_stats(dogs_shadow_dataset.name)
    assert stats.num_objects == 4
    assert stats.size == 15


def test_dataset_stats_registered_ds(cloud_test_catalog, dogs_registered_dataset):
    catalog = cloud_test_catalog.catalog
    stats = catalog.dataset_stats(dogs_registered_dataset.name, 1)
    assert stats.num_objects == 4
    assert stats.size == 15


def test_dataset_stats_missing_version(cloud_test_catalog, dogs_registered_dataset):
    catalog = cloud_test_catalog.catalog
    with pytest.raises(ValueError):
        catalog.dataset_stats(dogs_registered_dataset.name)



import os
import uuid
from pathlib import PosixPath
from typing import Dict, Generator, Type

import attrs
import pytest
from pytest import MonkeyPatch, TempPathFactory
from upath.implementations.cloud import CloudPath

from dql.catalog import Catalog
from dql.data_storage import SQLiteDataStorage
from dql.utils import DQLDir

from .utils import DEFAULT_TREE, instantiate_tree

# pylint: disable=redefined-outer-name,unused-argument

DEFAULT_DQL_BIN = "dql"
DEFAULT_DQL_GIT_REPO = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# JSON tests are currently ignored as they are a work in progress for SQLite
collect_ignore = ["setup.py", "unit/test_json.py"]


@pytest.fixture(scope="session")
def monkeypatch_session() -> Generator[MonkeyPatch, None, None]:
    """
    Like monkeypatch, but for session scope.
    """
    mpatch = pytest.MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(scope="session", autouse=True)
def clean_environment(
    monkeypatch_session: MonkeyPatch,  # pylint: disable=redefined-outer-name
    tmp_path_factory: TempPathFactory,
) -> None:
    """
    Make sure we have a clean environment and won't write to userspace.
    """
    working_dir = str(tmp_path_factory.mktemp("default_working_dir"))
    os.chdir(working_dir)
    monkeypatch_session.delenv(DQLDir.ENV_VAR, raising=False)

    def default_root(cls: Type[DQLDir]) -> str:
        return os.path.join(working_dir, cls.DEFAULT)

    monkeypatch_session.setattr(
        "dql.utils.DQLDir.default_root",
        classmethod(default_root),
    )


@pytest.fixture
def data_storage():
    _data_storage = SQLiteDataStorage(db_file=":memory:")
    yield _data_storage

    # Wipe the DB after the test
    # Using new connection to check that the DB isn't locked
    conn = _data_storage.clone().db
    tables = conn.execute(
        "SELECT name FROM sqlite_master WHERE type='table'"
    ).fetchall()
    for (table,) in tables:
        conn.execute(f"DROP TABLE '{table}'")
    conn.commit()


@pytest.fixture
def tmp_dir(tmp_path_factory, monkeypatch):
    dpath = tmp_path_factory.mktemp("dql-test")
    monkeypatch.chdir(dpath)
    return dpath


def pytest_addoption(parser):
    parser.addoption(
        "--dql-bin",
        type=str,
        default=DEFAULT_DQL_BIN,
        help="Path to dql binary",
    )

    parser.addoption(
        "--dql-revs",
        type=str,
        help="Comma-separated list of DQL revisions to test (overrides `--dql-bin`)",
    )

    parser.addoption(
        "--dql-git-repo",
        type=str,
        default=DEFAULT_DQL_GIT_REPO,
        help="Path or url to dql git repo",
    )
    parser.addoption(
        "--azure-connection-string",
        type=str,
        help=(
            "Connection string to run tests against a real, versioned "
            "Azure storage account"
        ),
    )


class DQLTestConfig:
    def __init__(self):
        self.dql_bin = DEFAULT_DQL_BIN
        self.dql_revs = None
        self.dql_git_repo = DEFAULT_DQL_GIT_REPO


@pytest.fixture(scope="session")
def test_config(request):
    return request.config.dql_config


def pytest_configure(config):
    config.dql_config = DQLTestConfig()

    config.dql_config.dql_bin = config.getoption("--dql-bin")
    config.dql_config.dql_revs = config.getoption("--dql-revs")
    config.dql_config.dql_git_repo = config.getoption("--dql-git-repo")


@pytest.fixture(scope="session", params=[DEFAULT_TREE])
def tree(request):
    return request.param


@attrs.define
class CloudServer:
    kind: str
    src: CloudPath
    client_config: Dict[str, str]


def make_cloud_server(src_path, cloud_type, tree):
    fs = src_path.fs
    if cloud_type == "s3":
        endpoint_url = fs.client_kwargs["endpoint_url"]
        client_config = {"aws_endpoint_url": endpoint_url}
    elif cloud_type == "gcs":
        endpoint_url = fs._endpoint  # pylint:disable=protected-access
        client_config = {"endpoint_url": endpoint_url}
    elif cloud_type == "azure":
        client_config = fs.storage_options.copy()
    elif cloud_type == "file":
        client_config = {}
    else:
        raise ValueError(f"invalid cloud_type: {cloud_type}")

    instantiate_tree(src_path, tree)
    return CloudServer(kind=cloud_type, src=src_path, client_config=client_config)


@attrs.define
class CloudTestCatalog:
    server: CloudServer
    working_dir: PosixPath
    catalog: Catalog

    @property
    def src(self):
        return self.server.src

    @property
    def src_uri(self):
        if self.server.kind == "file":
            return self.src.as_uri()
        return str(self.src).rstrip("/")

    @property
    def client_config(self):
        return self.server.client_config


@pytest.fixture(scope="session", params=["file", "s3", "gcs", "azure"])
def cloud_type(request):
    return request.param


@pytest.fixture(scope="session", params=[False, True])
def version_aware(request):
    return request.param


@pytest.fixture(scope="session")
def cloud_server(request, tmp_upath_factory, cloud_type, version_aware, tree):
    if cloud_type == "azure" and version_aware:
        if conn_str := request.config.getoption("--azure-connection-string"):
            src_path = tmp_upath_factory.azure(conn_str)
        else:
            pytest.skip("Can't test versioning with Azure")
    elif cloud_type == "file":
        if version_aware:
            pytest.skip("Local storage can't be versioned")
        else:
            src_path = tmp_upath_factory.mktemp("local")
    else:
        src_path = tmp_upath_factory.mktemp(cloud_type, version_aware=version_aware)
    return make_cloud_server(src_path, cloud_type, tree)


@pytest.fixture
def cloud_test_catalog(
    cloud_server,
    tmp_path,
    data_storage,
):
    cache_dir = tmp_path / ".dql" / "cache"
    cache_dir.mkdir(parents=True)
    tmpfile_dir = tmp_path / ".dql" / "tmp"
    tmpfile_dir.mkdir()

    catalog = Catalog(data_storage, cache_dir=str(cache_dir), tmp_dir=str(tmpfile_dir))
    result = CloudTestCatalog(
        server=cloud_server, working_dir=tmp_path, catalog=catalog
    )
    if cloud_server.kind == "file":
        catalog.add_storage(f"{result.src_uri}")
    return result


@pytest.fixture
def listed_bucket(cloud_test_catalog):
    list(
        cloud_test_catalog.catalog.ls(
            [cloud_test_catalog.src_uri],
            fields=["name"],
            client_config=cloud_test_catalog.client_config,
        )
    )


@pytest.fixture
def dogs_shadow_dataset(listed_bucket, cloud_test_catalog):
    shadow_dataset_name = uuid.uuid4().hex
    src_uri = cloud_test_catalog.src_uri
    catalog = cloud_test_catalog.catalog

    catalog.create_shadow_dataset(
        shadow_dataset_name,
        [f"{src_uri}/dogs/*"],
        client_config=cloud_test_catalog.client_config,
        recursive=True,
    )

    return catalog.data_storage.get_dataset(shadow_dataset_name)



import os
import os.path as osp
import stat
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
)

from dateutil import tz
from dateutil.parser import isoparse
from tomlkit import parse

if TYPE_CHECKING:
    from tomlkit import TOMLDocument

GLOB_CHARS = ["?", "*", "[", "]"]
NUL = b"\0"

T = TypeVar("T", bound="DQLDir")


class DQLDir:
    DEFAULT = ".dql"
    CACHE = "cache"
    TMP = "tmp"
    DB = "db"
    ENV_VAR = "DQL_DIR"

    def __init__(
        self,
        root: Optional[str] = None,
        cache: Optional[str] = None,
        tmp: Optional[str] = None,
        db: Optional[str] = None,
    ) -> None:
        self.root = osp.abspath(root) if root is not None else self.default_root()
        self.cache = (
            osp.abspath(cache) if cache is not None else osp.join(self.root, self.CACHE)
        )
        self.tmp = (
            osp.abspath(tmp) if tmp is not None else osp.join(self.root, self.TMP)
        )
        self.db = osp.abspath(db) if db is not None else osp.join(self.root, self.DB)

    def init(self):
        os.makedirs(self.root, exist_ok=True)
        os.makedirs(self.cache, exist_ok=True)
        os.makedirs(self.tmp, exist_ok=True)
        os.makedirs(osp.split(self.db)[0], exist_ok=True)

    @classmethod
    def default_root(cls) -> str:
        return osp.join(os.getcwd(), cls.DEFAULT)

    @classmethod
    def find(cls: Type[T], create: bool = True) -> T:
        try:
            root = os.environ[cls.ENV_VAR]
        except KeyError:
            root = cls.default_root()
        instance = cls(root)
        if not osp.isdir(root):
            if create:
                instance.init()
            else:
                NotADirectoryError(root)
        return instance


def read_config(dql_root: str) -> Optional["TOMLDocument"]:
    project_path = osp.join(dql_root, "config")
    if osp.isfile(project_path):
        with open(project_path, encoding="utf-8") as f:
            text = f.read()
        return parse(text)
    return None


def get_remote_config(
    config: Optional["TOMLDocument"], remote: str = ""
) -> Mapping[str, str]:
    if config is None:
        return {"type": "local"}
    if not remote:
        try:
            remote = config["core"]["default-remote"]  # type: ignore[index,assignment] # noqa: E501
        except KeyError:
            return {"type": "local"}
    try:
        remote_conf: Mapping[str, str] = config["remote"][remote]  # type: ignore[assignment,index] # noqa: E501
    except KeyError:
        raise Exception(f"missing config section for default remote: remote.{remote}")
    except Exception as exc:
        raise Exception("invalid config") from exc

    if not isinstance(remote_conf, Mapping):
        raise Exception(f"config section remote.{remote} must be a mapping")

    remote_type = remote_conf.get("type")
    if remote_type not in ("local", "http"):
        raise Exception(
            f'config section remote.{remote} must have "type" with one of: '
            '"local", "http"'
        )

    if remote_type == "http":
        for key in ["url", "username", "token"]:
            try:
                remote_conf[key]
            except KeyError:
                raise Exception(
                    f"config section remote.{remote} of type {remote_type} "
                    f"must contain key {key}"
                )
    elif remote_type != "local":
        raise Exception(
            f"config section remote.{remote} has invalid remote type {remote_type}"
        )
    return remote_conf


def human_time_to_int(time: str) -> Optional[int]:
    if not time:
        return None

    suffix = time[-1]
    try:
        num = int(time if suffix.isdigit() else time[:-1])
    except ValueError:
        return None
    return num * {
        "h": 60 * 60,
        "d": 60 * 60 * 24,
        "w": 60 * 60 * 24 * 7,
        "m": 31 * 24 * 60 * 60,
        "y": 60 * 60 * 24 * 365,
    }.get(suffix.lower(), 1)


def time_to_str(dt):
    if isinstance(dt, str):
        dt = isoparse(dt)
    return dt.strftime("%Y-%m-%d %H:%M:%S")


def time_to_local(dt: Union[datetime, str]) -> datetime:
    # TODO check usage
    if isinstance(dt, str):
        dt = isoparse(dt)
    try:
        return dt.astimezone(tz.tzlocal())
    except (OverflowError, OSError, ValueError):
        return dt


def time_to_local_str(dt: Union[datetime, str]) -> str:
    return time_to_str(time_to_local(dt))


def is_expired(expires: Optional[Union[datetime, str]]):
    if expires:
        return time_to_local(expires) < time_to_local(datetime.now())

    return False


SIZE_SUFFIXES = ["", "K", "M", "G", "T", "P", "E", "Z", "Y", "R", "Q"]


def sizeof_fmt(num, suffix="", si=False):
    power = 1000.0 if si else 1024.0
    for unit in SIZE_SUFFIXES[:-1]:
        if abs(num) < power:
            if not unit:
                return f"{num:4.0f}{suffix}"
            return f"{num:3.1f}{unit}{suffix}"
        num /= power
    return f"{num:.1f}Q{suffix}"


def suffix_to_number(num_str: str) -> int:
    try:
        if len(num_str) > 1:
            suffix = num_str[-1].upper()
            if suffix in SIZE_SUFFIXES:
                suffix_idx = SIZE_SUFFIXES.index(suffix)
                return int(num_str[:-1]) * (1024**suffix_idx)
        return int(num_str)
    except (TypeError, ValueError):
        raise ValueError(f"Invalid number/suffix for: {num_str}")


def force_create_dir(name):
    if not os.path.exists(name):
        os.mkdir(name)
    elif not os.path.isdir(name):
        os.remove(name)
        os.mkdir(name)


def dql_paths_join(source_path: str, file_paths: Iterable[str]) -> Iterable[str]:
    source_parts = source_path.rstrip("/").split("/")
    if set(source_parts[-1]).intersection(GLOB_CHARS):
        # Remove last element if it is a glob match (such as *)
        source_parts.pop()
    source_stripped = "/".join(source_parts)
    return (f"{source_stripped}/{path.lstrip('/')}" for path in file_paths)


# From: https://docs.python.org/3/library/shutil.html#rmtree-example
def remove_readonly(func, path, _):
    "Clear the readonly bit and reattempt the removal"
    os.chmod(path, stat.S_IWRITE)
    func(path)


def sql_escape_like(search: str, escape: str = "\\") -> str:
    return (
        search.replace(escape, escape * 2)
        .replace("%", f"{escape}%")
        .replace("_", f"{escape}_")
    )


def get_envs_by_prefix(prefix: str) -> Dict[str, str]:
    """
    Function that searches env variables by some name prefix and returns
    the ones found, but with prefix being excluded from it's names
    """
    variables: Dict[str, str] = {}
    for env_name, env_value in os.environ.items():
        if env_name.startswith(prefix):
            variables[env_name[len(prefix) :]] = env_value

    return variables



from dql.vendored.adlfs import AzureBlobFileSystem

from .fsspec import DELIMITER, FSSpecClient


class AzureClient(FSSpecClient):
    FS_CLASS = AzureBlobFileSystem
    PREFIX = "az://"
    protocol = "az"

    def url(self, path: str, expires: int = 3600) -> str:
        return self.fs.url(path, expires=expires)

    def _dict_from_info(self, v, parent_id, path, partial_id):
        version_id = v.get("version_id")
        checksum = v.get("content_settings", {}).get("content_md5", "")
        name = v.get("name", "").split(DELIMITER)[-1]
        if version_id:
            version_suffix = f"?versionid={version_id}"
            if name.endswith(version_suffix):
                name = name[: -len(version_suffix)]
        return {
            "dir_id": None,
            "parent_id": parent_id,
            "path": path,
            "name": name,
            "checksum": checksum or "",
            "etag": v.get("etag", ""),
            "version": version_id or "",
            "is_latest": version_id is None or bool(v.get("is_current_version")),
            "last_modified": v["last_modified"],
            "size": v.get("size", ""),
            "owner_name": "",
            "owner_id": "",
            "anno": None,
            "partial_id": partial_id,
        }



import json
import tarfile
from datetime import datetime, timezone
from random import getrandbits
from typing import TYPE_CHECKING, Any

from sqlalchemy import String

from dql.catalog import Catalog
from dql.client.checksum import file_digest
from dql.data_storage.abstract import RANDOM_BITS
from dql.node import DirType

from .udf import generator

if TYPE_CHECKING:
    from dql.dataset import DatasetRow


__all__ = ["index_tar", "checksum"]


def tarmember_from_info(info, parent):
    location = json.dumps(
        [
            {
                "offset": info.offset_data,
                "size": info.size,
                "type": "tar",
                "parent": parent.path_str,
            },
        ]
    )
    return {
        "vtype": "tar",
        "dir_type": DirType.FILE,
        "parent_id": parent.id,
        "path_str": f"{parent.path_str}/{info.name}",
        "name": info.name.rsplit("/", 1)[-1],
        "checksum": "",
        "etag": "",
        "version": "",
        "is_latest": parent.is_latest,
        "last_modified": datetime.fromtimestamp(info.mtime, timezone.utc),
        "size": info.size,
        "owner_name": info.uname,
        "owner_id": str(info.uid),
        "anno": None,
        "source": parent.source,
        "random": getrandbits(RANDOM_BITS),
        "location": location,
    }


@generator(Catalog)
def index_tar(row, catalog):
    with catalog.open_object(row) as f:
        with tarfile.open(fileobj=f, mode="r:") as archive:
            for info in archive:
                if info.isdir():
                    continue
                yield tarmember_from_info(info, parent=row)


class ChecksumFunc:
    """Calculate checksums for objects reference by dataset rows."""

    output_type = String

    column = "checksum"

    def __init__(self):
        pass

    def __call__(self, catalog: "Catalog", row: "DatasetRow") -> Any:
        with catalog.open_object(row) as f:
            return file_digest(f)


checksum = ChecksumFunc()



import json
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Optional, Sequence, Type, TypeVar

import attrs

T = TypeVar("T", bound="DatasetRecord")
V = TypeVar("V", bound="DatasetVersion")


@dataclass
class DatasetStats:
    num_objects: int
    size: int  # in bytes


class Status:
    CREATED = 1
    PENDING = 2
    FAILED = 3
    COMPLETE = 4
    STALE = 6


@dataclass
class DatasetVersion:
    id: int
    dataset_id: int
    version: int
    created_at: datetime

    @classmethod
    def parse(
        cls: Type[V],
        id: int,  # pylint: disable=redefined-builtin
        dataset_id: int,
        version: int,
        created_at: datetime,
    ):
        return cls(id, dataset_id, version, created_at)

    def __eq__(self, other):
        if not isinstance(other, DatasetVersion):
            return False
        return self.version == other.version and self.dataset_id == other.dataset_id

    def __lt__(self, other):
        if not isinstance(other, DatasetVersion):
            return False
        return self.version < other.version

    def __hash__(self):
        return hash(f"{self.dataset_id}_{self.version}")


@dataclass
class DatasetRecord:
    id: int
    name: str
    description: Optional[str]
    labels: Sequence[str]
    shadow: bool
    versions: Optional[List[DatasetVersion]]
    status: int = Status.CREATED
    created_at: Optional[datetime] = None
    finished_at: Optional[datetime] = None

    @classmethod
    def parse(
        cls: Type[T],
        id: int,  # pylint: disable=redefined-builtin
        name: str,
        description: Optional[str],
        labels: str,
        shadow: int,
        status: int,
        created_at: datetime,
        finished_at: Optional[datetime],
        version_id: Optional[int],
        version_dataset_id: Optional[int],
        version: Optional[int],
        version_created_at: Optional[datetime],
    ) -> "DatasetRecord":
        labels_lst: List[str] = json.loads(labels) if labels else []
        versions = None
        if version_id and version and version_dataset_id and version_created_at:
            versions = [
                DatasetVersion.parse(
                    version_id, version_dataset_id, version, version_created_at
                )
            ]

        return cls(
            id,
            name,
            description,
            labels_lst,
            bool(shadow),
            versions,
            status,
            created_at,
            finished_at,
        )

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def merge_versions(self, other: "DatasetRecord") -> "DatasetRecord":
        """Merge versions from another dataset"""
        if other.id != self.id:
            raise RuntimeError("Cannot merge versions of datasets with different ids")
        if not other.versions:
            # nothing to merge
            return self
        if not self.versions:
            self.versions = []

        self.versions = list(set(self.versions + other.versions))
        return self

    def sort_versions(self, reverse=False) -> None:
        """Sorts versions by version number"""
        if not self.versions:
            return
        self.versions.sort(key=lambda v: v.version, reverse=reverse)

    def has_version(self, version: int) -> bool:
        return version in self.versions_values

    def is_valid_next_version(self, version: int) -> bool:
        """
        Checks if a number can be a valid next latest version for dataset.
        The only rule is that it cannot be lower than current latest version
        """
        if self.latest_version and self.latest_version >= version:
            return False
        return True

    def remove_version(self, version: int) -> None:
        if not self.versions or not self.has_version(version):
            return

        self.versions = [v for v in self.versions if v.version != version]

    @property
    def registered(self) -> bool:
        return not self.shadow

    @property
    def versions_values(self) -> List[int]:
        """
        Extracts actual versions from list of DatasetVersion objects
        in self.versions attribute
        """
        if not self.versions:
            return []

        return sorted([v.version for v in self.versions])

    @property
    def next_version(self) -> int:
        """Returns what should be next autoincrement version of dataset"""
        if self.shadow or not self.versions:
            return 1
        return max(self.versions_values) + 1

    @property
    def latest_version(self) -> Optional[int]:
        """Returns latest version of a dataset"""
        if self.shadow or not self.versions:
            return None
        return max(self.versions_values)

    @property
    def prev_version(self) -> Optional[int]:
        """Returns previous version of a dataset"""
        if self.shadow or not self.versions or len(self.versions) == 1:
            return None

        return sorted(self.versions_values)[-2]


@attrs.define
class DatasetRow:
    id: int
    vtype: str
    dir_type: int
    parent_id: Optional[int]
    name: str
    checksum: str
    etag: str
    version: str
    is_latest: bool
    last_modified: Optional[datetime]
    size: int
    owner_name: str
    owner_id: str
    path_str: str
    anno: Optional[str]
    random: int
    location: Optional[str]
    source: str
    custom: Optional[Dict] = attrs.field(factory=dict)

    @classmethod
    def from_result_row(cls, columns, values):
        row = dict(zip(columns, values))
        core_fields = {key: value for (key, value) in row.items() if hasattr(cls, key)}
        custom_fields = {
            key: value for (key, value) in row.items() if key not in core_fields
        }
        return cls(**core_fields, custom=custom_fields)

    def __getitem__(self, col):
        if hasattr(self, col):
            return getattr(self, col)
        elif self.custom and col in self.custom:
            return self.custom[col]
        raise KeyError



from datetime import datetime
from typing import Any, Dict, List, Optional

import attrs

from dql.utils import time_to_str


class DirType:
    FILE = 0
    DIR = 1
    TAR_ARCHIVE = 5


class DirTypeGroup:
    """
    Groups of DirTypes for selecting storage nodes or dataset entries.

    When filtering with FILE and DIR together or alternatively when
    using SUBOBJ_FILE and SUBOBJ_DIR together, we achieve a
    filesystem-compatible view of a storage location. Such a view
    avoids path conflicts and could be downloaded as a directory tree.

    FILE, DIR
      The respective types which appear on the indexed filesystem or
      object store as a file or directory.

    SUBOBJ_FILE, SUBOBJ_DIR
      The respective types that we want to consider to be a file or
      directory when including subobjects which are generated from other
      files. In this case, we treat tar archives as directories so tar
      subobjects (TAR_FILE) can be viewed under the directory tree of
      the parent tar archive.
    """

    FILE = (DirType.FILE, DirType.TAR_ARCHIVE)
    DIR = (DirType.DIR,)
    SUBOBJ_FILE = (DirType.FILE,)
    SUBOBJ_DIR = (DirType.DIR, DirType.TAR_ARCHIVE)


@attrs.define
class Node:
    id: int = 0
    vtype: str = ""
    dir_type: Optional[int] = None
    parent_id: Optional[int] = None
    name: str = ""
    checksum: str = ""
    etag: str = ""
    version: Optional[str] = None
    is_latest: bool = True
    last_modified: Optional[datetime] = None
    size: int = 0
    owner_name: str = ""
    owner_id: str = ""
    path_str: str = ""
    anno: Optional[str] = None
    random: int = -1
    location: Optional[str] = None
    partial_id: int = 0
    valid: bool = True

    @property
    def basename(self):
        # Note: differs from self.name only for TAR sub-objects
        return self.name.rsplit("/", 1)[-1]

    @property
    def is_dir(self) -> bool:
        return self.dir_type == DirType.DIR

    @property
    def is_container(self) -> bool:
        return self.dir_type in DirTypeGroup.SUBOBJ_DIR

    @property
    def is_downloadable(self) -> bool:
        return bool(not self.is_dir and self.name)

    def sql_schema(self):
        return ",".join(["?"] * len(self))

    def append_to_file(self, fd, path: str):
        fd.write(f"- name: {path}\n")
        fd.write(f"  etag: {self.etag}\n")
        checksum = self.checksum
        if checksum:
            fd.write(f"  checksum: {self.checksum}\n")
        version = self.version
        if version:
            fd.write(f"  version: {self.version}\n")
        fd.write(f"  last_modified: '{time_to_str(self.last_modified)}'\n")
        size = self.size
        fd.write(f"  size: {self.size}\n")
        return size

    def get_metafile_data(self, path: str):
        data: Dict[str, Any] = {
            "name": path,
            "etag": self.etag,
        }
        checksum = self.checksum
        if checksum:
            data["checksum"] = checksum
        version = self.version
        if version:
            data["version"] = version
        data["last_modified"] = time_to_str(self.last_modified)
        data["size"] = self.size
        return data

    @property
    def full_path(self) -> str:
        if self.is_dir and self.path_str:
            return self.path_str + "/"
        return self.path_str


@attrs.define
class NodeWithPath:
    n: Node
    path: List[str] = []

    def append_to_file(self, fd):
        return self.n.append_to_file(fd, "/".join(self.path))

    def get_metafile_data(self):
        return self.n.get_metafile_data("/".join(self.path))

    @property
    def full_path(self) -> str:
        path = "/".join(self.path)
        if self.n.is_dir and path:
            path += "/"
        return path


TIME_FMT = "%Y-%m-%d %H:%M"


def long_line_str(name: str, timestamp: Optional[datetime], owner: str) -> str:
    if timestamp is None:
        time = "-"
    else:
        time = timestamp.strftime(TIME_FMT)
    return f"{owner: <19} {time: <19} {name}"



import logging
import os
import shlex
import sys
import traceback
from argparse import Action, ArgumentParser, ArgumentTypeError, Namespace
from itertools import chain
from typing import Iterable, Iterator, List, Mapping, Optional, Tuple, Union

import shtab

from dql import __version__, utils
from dql.catalog import Catalog, get_catalog, indexer_formats
from dql.cli_utils import BooleanOptionalAction
from dql.client import Client
from dql.node import DirType, long_line_str
from dql.remote.studio import StudioClient
from dql.utils import DQLDir, get_remote_config, read_config

logger = logging.getLogger("dql")

TTL_HUMAN = "4h"
TTL_INT = 4 * 60 * 60
FIND_COLUMNS = ["du", "name", "owner", "path", "size", "type"]


def human_time_type(value_str: str, can_be_none: bool = False) -> Optional[int]:
    value = utils.human_time_to_int(value_str)

    if value:
        return value
    if can_be_none:
        return None

    raise ArgumentTypeError(
        "This option supports only a human-readable time interval like 12h or 4w."
    )


def parse_find_column(column: str) -> str:
    column_lower = column.strip().lower()
    if column_lower in FIND_COLUMNS:
        return column_lower
    raise ArgumentTypeError(
        f"Invalid column for find: '{column}' Options are: {','.join(FIND_COLUMNS)}"
    )


def find_columns_type(
    columns_str: str,
    default_colums_str: str = "path",
) -> List[str]:
    if not columns_str:
        columns_str = default_colums_str

    return [parse_find_column(c) for c in columns_str.split(",")]


def add_sources_arg(parser: ArgumentParser, nargs: Union[str, int] = "+") -> Action:
    return parser.add_argument(
        "sources",
        type=str,
        nargs=nargs,
        help="Data sources - paths to cloud storage dirs",
    )


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description="DQL: Data Query Language", prog="dql")

    parser.add_argument("-V", "--version", action="version", version=__version__)

    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "--aws-endpoint-url",
        type=str,
        help="AWS endpoint URL",
    )
    parent_parser.add_argument(
        "--aws-anon",
        action="store_true",
        help="AWS anon (aka awscli's --no-sign-request)",
    )
    parent_parser.add_argument(
        "--ttl",
        type=human_time_type,
        default=TTL_HUMAN,
        help="Time-to-live of data source cache. Negative equals forever.",
    )
    parent_parser.add_argument(
        "-u", "--update", action="count", default=0, help="Update cache"
    )
    parent_parser.add_argument(
        "-v", "--verbose", action="count", default=0, help="Verbose"
    )
    parent_parser.add_argument(
        "-q", "--quiet", action="count", default=0, help="Be quiet"
    )

    subp = parser.add_subparsers(help="Sub-command help", dest="command")
    parse_cp = subp.add_parser(
        "cp", parents=[parent_parser], help="Copy data files from the cloud"
    )
    add_sources_arg(parse_cp).complete = shtab.DIR  # type: ignore[attr-defined] # noqa: E501
    parse_cp.add_argument("output", type=str, help="Output")
    parse_cp.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Force creating outputs",
    )
    parse_cp.add_argument(
        "-r",
        "-R",
        "--recursive",
        default=False,
        action="store_true",
        help="Copy directories recursively",
    )
    parse_cp.add_argument(
        "--no-glob",
        default=False,
        action="store_true",
        help="Do not expand globs (such as * or ?)",
    )

    parse_clone = subp.add_parser(
        "clone", parents=[parent_parser], help="Copy data files from the cloud"
    )
    add_sources_arg(parse_clone).complete = shtab.DIR  # type: ignore[attr-defined] # noqa: E501
    parse_clone.add_argument("output", type=str, help="Output")
    parse_clone.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Force creating outputs",
    )
    parse_clone.add_argument(
        "-r",
        "-R",
        "--recursive",
        default=False,
        action="store_true",
        help="Copy directories recursively",
    )
    parse_clone.add_argument(
        "--no-glob",
        default=False,
        action="store_true",
        help="Do not expand globs (such as * or ?)",
    )
    parse_clone.add_argument(
        "--no-cp",
        default=False,
        action="store_true",
        help="Do not copy files, just create a shadow dataset",
    )
    parse_clone.add_argument(
        "--edql",
        default=False,
        action="store_true",
        help="Create a .edql file",
    )
    parse_clone.add_argument(
        "--edql-file",
        help="Use a different filename for the resulting .edql file",
    )

    parse_register = subp.add_parser(
        "register", parents=[parent_parser], help="Register a dataset"
    )
    parse_register.add_argument("name", type=str, help="Dataset name")
    parse_register.add_argument(
        "--registered-name",
        action="store",
        default="",
        help="Registered dataset name",
    )
    parse_register.add_argument(
        "--version",
        action="store",
        default=None,
        type=int,
        help="New registered dataset version",
    )
    parse_register.add_argument(
        "--description",
        action="store",
        default="",
        help="Dataset description",
    )
    parse_register.add_argument(
        "--labels",
        default=[],
        nargs="+",
        help="Dataset labels",
    )

    parse_edit_dataset = subp.add_parser(
        "edit-dataset", parents=[parent_parser], help="Edit dataset metadata"
    )
    parse_edit_dataset.add_argument("name", type=str, help="Dataset name")
    parse_edit_dataset.add_argument(
        "--new-name",
        action="store",
        default="",
        help="Dataset new name",
    )
    parse_edit_dataset.add_argument(
        "--description",
        action="store",
        default="",
        help="Dataset description",
    )
    parse_edit_dataset.add_argument(
        "--labels",
        default=[],
        nargs="+",
        help="Dataset labels",
    )

    ls_datasets_parser = subp.add_parser(
        "ls-datasets", parents=[parent_parser], help="List datasets"
    )
    ls_datasets_parser.add_argument(
        "--shadow",
        default=None,
        action=BooleanOptionalAction,
        help="List only shadow datasets, or only registered datasets",
    )

    ls_dataset_rows_parser = subp.add_parser(
        "ls-dataset-rows", parents=[parent_parser], help="List dataset rows"
    )
    ls_dataset_rows_parser.add_argument("name", type=str, help="Dataset name")
    ls_dataset_rows_parser.add_argument(
        "--version",
        action="store",
        default=None,
        type=int,
        help="Dataset version",
    )

    rm_dataset_parser = subp.add_parser(
        "rm-dataset", parents=[parent_parser], help="Removes dataset"
    )
    rm_dataset_parser.add_argument("name", type=str, help="Dataset name")
    rm_dataset_parser.add_argument(
        "--version",
        action="store",
        default=None,
        type=int,
        help="Dataset version",
    )
    rm_dataset_parser.add_argument(
        "--force",
        default=False,
        action=BooleanOptionalAction,
        help="Force delete registered dataset with all of it's versions",
    )

    dataset_stats_parser = subp.add_parser(
        "dataset-stats", parents=[parent_parser], help="Shows basic dataset stats"
    )
    dataset_stats_parser.add_argument("name", type=str, help="Dataset name")
    dataset_stats_parser.add_argument(
        "--version",
        action="store",
        default=None,
        type=int,
        help="Dataset version",
    )
    dataset_stats_parser.add_argument(
        "-b",
        "--bytes",
        default=False,
        action="store_true",
        help="Display size in bytes instead of human-readable size",
    )
    dataset_stats_parser.add_argument(
        "--si",
        default=False,
        action="store_true",
        help="Display size using powers of 1000 not 1024",
    )

    parse_merge_datasets = subp.add_parser(
        "merge-datasets", parents=[parent_parser], help="Merges datasets"
    )
    parse_merge_datasets.add_argument(
        "--src",
        action="store",
        default=None,
        help="Source dataset name",
    )
    parse_merge_datasets.add_argument(
        "--dst",
        action="store",
        default=None,
        help="Destination dataset name",
    )
    parse_merge_datasets.add_argument(
        "--src-version",
        action="store",
        default=None,
        type=int,
        help="Source dataset version",
    )
    parse_merge_datasets.add_argument(
        "--dst-version",
        action="store",
        default=None,
        type=int,
        help="Destination dataset version",
    )

    parse_ls = subp.add_parser(
        "ls", parents=[parent_parser], help="List storage contents"
    )
    add_sources_arg(parse_ls, nargs="*")
    parse_ls.add_argument(
        "-l",
        "--long",
        action="count",
        default=0,
        help="List files in the long format",
    )
    parse_ls.add_argument(
        "--remote",
        action="store",
        default="",
        help="Name of remote to use",
    )

    parse_du = subp.add_parser(
        "du", parents=[parent_parser], help="Display space usage"
    )
    add_sources_arg(parse_du)
    parse_du.add_argument(
        "-b",
        "--bytes",
        default=False,
        action="store_true",
        help="Display sizes in bytes instead of human-readable sizes",
    )
    parse_du.add_argument(
        "-d",
        "--depth",
        "--max-depth",
        default=0,
        type=int,
        metavar="N",
        help=(
            "Display sizes for N directory depths below the given directory, "
            "the default is 0 (summarize provided directory only)."
        ),
    )
    parse_du.add_argument(
        "--si",
        default=False,
        action="store_true",
        help="Display sizes using powers of 1000 not 1024",
    )

    parse_find = subp.add_parser(
        "find", parents=[parent_parser], help="Search in a directory hierarchy"
    )
    add_sources_arg(parse_find)
    parse_find.add_argument(
        "-name",
        "--name",
        type=str,
        action="append",
        help="Filename to match pattern.",
    )
    parse_find.add_argument(
        "-iname",
        "--iname",
        type=str,
        action="append",
        help="Like -name but case insensitive.",
    )
    parse_find.add_argument(
        "-path",
        "--path",
        type=str,
        action="append",
        help="Path to match pattern.",
    )
    parse_find.add_argument(
        "-ipath",
        "--ipath",
        type=str,
        action="append",
        help="Like -path but case insensitive.",
    )
    parse_find.add_argument(
        "-size",
        "--size",
        type=str,
        help=(
            "Filter by size (+ is greater or equal, - is less or equal). "
            "Specified size is in bytes, or use a suffix like K, M, G for "
            "kilobytes, megabytes, gigabytes, etc."
        ),
    )
    parse_find.add_argument(
        "-type",
        "--type",
        type=str,
        help='File type: "f" - regular, "d" - directory',
    )
    parse_find.add_argument(
        "-jmespath",
        "--jmespath",
        type=str,
        action="append",
        help="JMESPath query to annotation (not supported currently)",
    )
    parse_find.add_argument(
        "-c",
        "--columns",
        type=find_columns_type,
        default=None,
        help=(
            "A comma-separated list of columns to print for each result. "
            f"Options are: {','.join(FIND_COLUMNS)} (Default: path)"
        ),
    )

    parse_index = subp.add_parser(
        "index", parents=[parent_parser], help="Index storage location"
    )
    parse_index.add_argument(
        "--format",
        type=str,
        default=None,
        choices=indexer_formats.keys(),
        help="Postprocessing step to apply, after indexing storage.",
    )
    add_sources_arg(parse_index)
    add_storage_parser = subp.add_parser(
        "add-storage", parents=[parent_parser], help="Register local path as storage"
    )
    add_storage_parser.add_argument("uri", type=str, help="Source URI or path")
    add_storage_parser.add_argument(
        "--symlinks",
        default=False,
        action="store_true",
        help="Use symlinks instead of full copies when downloading from storage",
    )

    subp.add_parser(
        "find-stale-storages",
        parents=[parent_parser],
        help="Finds and marks stale storages",
    )

    add_completion_parser(subp, [parent_parser])
    return parser


def add_completion_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "completion",
        parents=parents,
        help="Output shell completion script",
    )
    parser.add_argument(
        "-s",
        "--shell",
        help="Shell syntax for completions.",
        default="bash",
        choices=shtab.SUPPORTED_SHELLS,
    )


def get_logging_level(args: Namespace) -> int:
    if args.quiet:
        return logging.CRITICAL
    elif args.verbose:
        return logging.DEBUG
    return logging.INFO


def ls_urls(
    sources,
    long: bool = False,
    *,
    client_config=None,
    catalog: Optional[Catalog] = None,
    **kwargs,
) -> Iterator[Tuple[str, Iterator[str]]]:
    if client_config is None:
        client_config = {}
    if catalog is None:
        catalog = get_catalog()

    curr_dir = None
    value_iterables = []
    for next_dir, values in _ls_urls_flat(
        sources, long, client_config, catalog, **kwargs
    ):
        if curr_dir is None or next_dir == curr_dir:  # type: ignore[unreachable] # noqa: E501
            value_iterables.append(values)
        else:
            yield curr_dir, chain(*value_iterables)  # type: ignore[unreachable] # noqa: E501
            value_iterables = [values]
        curr_dir = next_dir
    if curr_dir is not None:
        yield curr_dir, chain(*value_iterables)


def _node_data_to_ls_values(row, long_format=False):
    name = row[0]
    is_dir = row[1] == DirType.DIR
    ending = "/" if is_dir else ""
    value = name + ending
    if long_format:
        last_modified = row[2]
        owner_name = row[3]
        timestamp = last_modified if not is_dir else None
        return long_line_str(value, timestamp, owner_name)
    return value


def _ls_urls_flat(
    sources,
    long: bool,
    client_config,
    catalog: Catalog,
    **kwargs,
) -> Iterator[Tuple[str, Iterator[str]]]:
    for source in sources:
        client_cls = Client.get_implementation(source)
        if client_cls.is_root_url(source):
            buckets = client_cls.ls_buckets(**client_config)
            if long:
                values = (long_line_str(b.name, b.created, "") for b in buckets)
            else:
                values = (b.name for b in buckets)
            yield source, values
        else:
            found = False
            fields = ["name", "dir_type"]
            if long:
                fields.extend(["last_modified", "owner_name"])
            for data_source, results in catalog.ls(
                [source], fields=fields, client_config=client_config, **kwargs
            ):
                values = (_node_data_to_ls_values(r, long) for r in results)
                found = True
                yield data_source.dirname(), values
            if not found:
                raise FileNotFoundError(f"No such file or directory: {source}")


def ls_indexed_storages(
    long: bool = False, catalog: Optional[Catalog] = None
) -> Iterator[str]:
    if catalog is None:
        catalog = get_catalog()
    storage_uris = catalog.ls_storage_uris()
    if long:
        for uri in storage_uris:
            # TODO: add Storage.created so it can be used here
            yield long_line_str(uri, None, "")
    else:
        yield from storage_uris


def ls_local(sources, long: bool = False, catalog: Optional[Catalog] = None, **kwargs):
    if catalog is None:
        catalog = get_catalog()
    if sources:
        if len(sources) == 1:
            for _, entries in ls_urls(sources, long=long, catalog=catalog, **kwargs):
                for entry in entries:
                    print(format_ls_entry(entry))
        else:
            first = True
            for source, entries in ls_urls(
                sources, long=long, catalog=catalog, **kwargs
            ):
                # print a newline between directory listings
                if first:
                    first = False
                else:
                    print()
                if source:
                    print(f"{source}:")
                for entry in entries:
                    print(format_ls_entry(entry))
    else:
        for entry in ls_indexed_storages(long=long, catalog=catalog):
            print(format_ls_entry(entry))


def format_ls_entry(entry: str) -> str:
    if entry.endswith("/") or not entry:
        entry = shlex.quote(entry[:-1])
        return f"{entry}/"
    return shlex.quote(entry)


def ls_remote(
    url: str,
    username: str,
    token: str,
    paths: Iterable[str],
    long: bool = False,
):
    client = StudioClient(url, username, token)
    first = True
    for path, response in client.ls(paths):
        if not first:
            print()
        if not response.ok or response.data is None:
            print(f"{path}:\n  Error: {response.message}\n")
            continue

        print(f"{path}:")
        if long:
            for row in response.data:
                entry = long_line_str(
                    row["name"] + ("/" if row["dir_type"] else ""),
                    row["last_modified"],
                    row["owner_name"],
                )
                print(format_ls_entry(entry))
        else:
            for row in response.data:
                entry = row["name"] + ("/" if row["dir_type"] else "")
                print(format_ls_entry(entry))
        first = False


def ls(
    sources,
    long: bool = False,
    remote: str = "",
    config: Optional[Mapping[str, str]] = None,
    **kwargs,
):
    if config is None:
        config = get_remote_config(read_config(DQLDir.find().root), remote=remote)
    remote_type = config["type"]
    if remote_type == "local":
        ls_local(sources, long=long, **kwargs)
    else:
        ls_remote(
            config["url"],
            config["username"],
            config["token"],
            sources,
            long=long,
        )


def ls_datasets(shadow_only=None):
    for d in get_catalog().ls_datasets(shadow_only=shadow_only):
        if d.shadow:
            print(f"(tmp) {d.name}")
        else:
            for v in d.versions or []:
                print(f"{d.name} (v{v.version})")


def ls_dataset_rows(name: str, version: Optional[int] = None):
    for row in get_catalog().ls_dataset_rows(name, version=version):
        entry = row.name + ("/" if row.dir_type else "")  # type: ignore[attr-defined]
        print(format_ls_entry(entry))


def rm_dataset(name: str, version: Optional[int] = None, force: Optional[bool] = False):
    get_catalog().remove_dataset(name, version=version, force=force)


def dataset_stats(
    name: str,
    version: Optional[int] = None,
    show_bytes=False,
    si=False,
):
    stats = get_catalog().dataset_stats(name, version=version)

    if stats:
        print(f"Number of objects: {stats.num_objects}")
        if show_bytes:
            print(f"Total objects size: {stats.size}")
        else:
            print(f"Total objects size: {utils.sizeof_fmt(stats.size, si=si): >7}")


def du(sources, show_bytes=False, si=False, **kwargs):
    for path, size in get_catalog().du(sources, **kwargs):
        if show_bytes:
            print(f"{size} {path}")
        else:
            print(f"{utils.sizeof_fmt(size, si=si): >7} {path}")


def find(sources, **kwargs):
    yield from get_catalog().find(sources, **kwargs)


def index(
    sources,
    catalog: Optional[Catalog] = None,
    indexer_format: Optional[str] = None,
    **kwargs,
):
    if catalog is None:
        catalog = get_catalog()
    fmt = None
    if indexer_format:
        fmt = indexer_formats.get(indexer_format)
    catalog.index(sources, index_processors=fmt, **kwargs)


def add_storage(uri: str, symlinks: bool = False) -> None:
    get_catalog().add_storage(uri, symlinks=symlinks)


def find_stale_storages():
    get_catalog().find_stale_storages()


def completion(shell: str) -> str:
    return shtab.complete(
        get_parser(),
        shell=shell,  # nosec B604
    )


def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
    parser = get_parser()
    args = parser.parse_args(argv)

    if args.command is None:
        parser.print_help()
        return 1

    logger.addHandler(logging.StreamHandler())
    logging_level = get_logging_level(args)
    logger.setLevel(logging_level)

    client_config = {
        "aws_endpoint_url": args.aws_endpoint_url,
        "aws_anon": args.aws_anon,
    }

    try:
        if args.command == "cp":
            get_catalog().cp(
                args.sources,
                args.output,
                force=bool(args.force),
                update=bool(args.update),
                recursive=bool(args.recursive),
                edql_file=None,
                edql_only=False,
                no_edql_file=True,
                no_glob=args.no_glob,
                ttl=args.ttl,
                client_config=client_config,
            )
        elif args.command == "clone":
            get_catalog().clone(
                args.sources,
                args.output,
                force=bool(args.force),
                update=bool(args.update),
                recursive=bool(args.recursive),
                no_glob=args.no_glob,
                ttl=args.ttl,
                no_cp=args.no_cp,
                edql=args.edql,
                edql_file=args.edql_file,
                client_config=client_config,
            )
        elif args.command == "register":
            get_catalog().register_shadow_dataset(
                args.name,
                description=args.description,
                labels=args.labels,
                registered_name=args.registered_name,
                version=args.version,
            )
            print(f"Registered dataset {args.name}")
        elif args.command == "edit-dataset":
            get_catalog().edit_dataset(
                args.name,
                description=args.description,
                new_name=args.new_name,
                labels=args.labels,
            )
        elif args.command == "merge-datasets":
            get_catalog().merge_datasets(
                args.src,
                args.dst,
                src_version=args.src_version,
                dst_version=args.dst_version,
            )
        elif args.command == "ls":
            ls(
                args.sources,
                long=bool(args.long),
                remote=args.remote,
                ttl=args.ttl,
                update=bool(args.update),
                client_config=client_config,
            )
        elif args.command == "ls-datasets":
            ls_datasets(args.shadow)
        elif args.command == "ls-dataset-rows":
            ls_dataset_rows(args.name, args.version)
        elif args.command == "rm-dataset":
            rm_dataset(args.name, version=args.version, force=args.force)
        elif args.command == "dataset-stats":
            dataset_stats(
                args.name,
                version=args.version,
                show_bytes=args.bytes,
                si=args.si,
            )
        elif args.command == "du":
            du(
                args.sources,
                show_bytes=args.bytes,
                depth=args.depth,
                si=args.si,
                ttl=args.ttl,
                update=bool(args.update),
                client_config=client_config,
            )
        elif args.command == "find":
            results_found = False
            for result in find(
                args.sources,
                ttl=args.ttl,
                update=bool(args.update),
                names=args.name,
                inames=args.iname,
                paths=args.path,
                ipaths=args.ipath,
                size=args.size,
                typ=args.type,
                jmespath=args.jmespath,
                columns=args.columns,
                client_config=client_config,
            ):
                print(result)
                results_found = True
            if not results_found:
                print("No results")
        elif args.command == "index":
            index(
                args.sources,
                ttl=args.ttl,
                indexer_format=args.format,
                update=bool(args.update),
                client_config=client_config,
            )
        elif args.command == "add-storage":
            add_storage(args.uri, symlinks=args.symlinks)
        elif args.command == "completion":
            print(completion(args.shell))
        elif args.command == "find-stale-storages":
            find_stale_storages()
        else:
            print(f"invalid command: {args.command}", file=sys.stderr)
            return 1
        return 0
    except BrokenPipeError:
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        # See: https://docs.python.org/3/library/signal.html#note-on-sigpipe
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        return 141  # 128 + 13 (SIGPIPE)
    except Exception as exc:  # pylint: disable=broad-except
        print("Error:", exc, file=sys.stderr)
        if logging_level <= logging.DEBUG:
            traceback.print_exception(
                type(exc),
                exc,
                exc.__traceback__,
                file=sys.stderr,
            )
        return 1



import os
import random
import string
from abc import ABC, abstractmethod
from copy import copy
from typing import TYPE_CHECKING, Iterable, List, Optional, TypeVar

import sqlalchemy
from sqlalchemy.sql.expression import label

from dql.catalog import get_catalog
from dql.data_storage.schema import DATASET_CORE_COLUMN_NAMES
from dql.dataset import DatasetRow
from dql.dataset import Status as DatasetStatus

from .schema import C, create_udf_table
from .udf import UDFType

if TYPE_CHECKING:
    from sqlalchemy.sql.base import Executable
    from sqlalchemy.sql.elements import ColumnElement

    from dql.catalog import Catalog
    from dql.query.udf import Generator


BATCH_SIZE = 1000


class StartingStep(ABC):
    """An initial query processing step, referencing a data source."""

    @abstractmethod
    def apply(self) -> "QueryGenerator":
        ...


class Step(ABC):
    """A query processing step (filtering, mutation, etc.)"""

    @abstractmethod
    def apply(self, query: "Executable") -> "QueryGenerator":
        """Apply the processing step."""


class QueryStep(StartingStep):
    def __init__(self, table):
        self.table = table

    def apply(self):
        """Return the query for the table the query refers to."""
        table = self.table

        def q(*columns):
            return sqlalchemy.select(*columns)

        return QueryGenerator(q, table.c)


class IndexingStep(StartingStep):
    def __init__(self, path, catalog, **kwargs):
        self.path = path
        self.catalog = catalog
        self.kwargs = kwargs

    def apply(self):
        """Return the query for the table the query refers to."""
        self.catalog.index([self.path], **self.kwargs)
        client_config = self.kwargs.get("client_config") or {}
        client, path = self.catalog.parse_url(self.path, **client_config)
        uri = client.uri

        def q(*columns):
            col_names = [c.name for c in columns]
            return self.catalog.data_storage.nodes_dataset_query(
                column_names=col_names, path=path, recursive=True, uri=uri
            )

        column_objs = [sqlalchemy.column(c) for c in DATASET_CORE_COLUMN_NAMES]
        return QueryGenerator(q, column_objs)


class QueryGenerator:
    def __init__(self, func, columns):
        self.func = func
        self.columns = columns

    def exclude(self, column_names):
        return self.func(*[c for c in self.columns if c.name not in column_names])

    def select(self, column_names=None):
        if column_names is None:
            return self.func(*self.columns)
        return self.func(*[c for c in self.columns if c.name in column_names])


class UDFSignal(Step):
    """Add a custom column to the result set."""

    def __init__(self, name: str, udf: UDFType, catalog: "Catalog"):
        self.catalog = catalog
        self.name = name
        self.udf = udf

    def clone(self):
        return self.__class__(self.name, self.udf, self.catalog)

    def apply(self, query):
        if self.name in DATASET_CORE_COLUMN_NAMES:
            # populating a core dataset column
            return self.overwrite_existing(query)
        return self.new_column(query)

    def overwrite_existing(self, query):
        """Handler for udfs populating core columns (e.g. the checksum column)."""
        tbl, col = self.udf_results_table(query)
        # Construct a new query that will join the udf-generated partial table.

        subquery = query.subquery()
        q_cols = [c if c.name != self.name else col for c in subquery.c]

        def q(*columns):
            cols = []
            for c in columns:
                if c.name == col.name:
                    cols.append(col)
                else:
                    cols.append(c)

            q = subquery.join(tbl, tbl.c.id == subquery.c.id)
            if cols:
                return sqlalchemy.select(*cols).select_from(q)
            return sqlalchemy.select(cols).select_from(q)

        return QueryGenerator(q, q_cols)

    def new_column(self, query):
        """Handler for udfs adding custom signals."""
        tbl, col = self.udf_results_table(query)
        # Construct a new query that will join the udf-generated partial table.
        subq = query.subquery()

        def q(*columns):
            cols1 = []
            cols2 = []
            for c in columns:
                if c.name == col.name:
                    cols2.append(c)
                else:
                    cols1.append(c)

            if cols2:
                return (
                    sqlalchemy.select(*cols1)
                    .select_from(subq)
                    .join(tbl, tbl.c.id == subq.c.id)
                    .add_columns(*cols2)
                )
            return sqlalchemy.select(*cols1).select_from(subq)

        return QueryGenerator(q, [*subq.c, col])

    def udf_results_table(self, query):
        """
        Create and populate a temporary UDF results table, this table
        will have two columns:
         - id, for joining to the original query
         - {self.name}, for storing UDF results.
        """
        execute = self.catalog.data_storage.execute

        temp_table_name = f"udf_{self.name}_" + _random_string(6)
        col = sqlalchemy.Column(self.name, self.udf.output_type, nullable=True)
        tbl = create_udf_table(
            execute,
            temp_table_name,
            [col],
        )
        selected_columns = [col.name for col in query.selected_columns]
        results = (
            DatasetRow.from_result_row(selected_columns, r) for r in execute(query)
        )
        rows = []
        for row in results:
            signal = {self.name: self.udf(self.catalog, row)}
            rows.append(dict(id=row.id, **signal))
            if len(rows) > BATCH_SIZE:
                update = tbl.insert().values(rows)
                execute(update)
                rows.clear()
        if rows:
            update = tbl.insert().values(rows)
            execute(update)
        return tbl, col


class RowGenerator(Step):
    """Extend dataset with new rows."""

    def __init__(self, generator: "Generator", catalog: "Catalog"):
        self.generator = generator
        self.catalog = catalog
        self.engine = catalog.data_storage.engine

    def clone(self):
        return self.__class__(self.generator, self.catalog)

    def apply(self, query):
        # Create a temporary table.
        temp_table_name = "generated_" + _random_string(6)
        selected_columns = [col.name for col in query.selected_columns]
        custom_columns: List["sqlalchemy.Column"] = [
            sqlalchemy.Column(col.name, col.type)
            for col in query.selected_columns
            if col.name not in DATASET_CORE_COLUMN_NAMES
        ]
        self.catalog.data_storage.create_dataset_rows_table(
            temp_table_name,
            custom_columns=custom_columns,
            if_not_exists=False,
        )
        tbl = self.catalog.data_storage.get_table(temp_table_name)

        execute = self.catalog.data_storage.execute

        results = (
            DatasetRow.from_result_row(selected_columns, r) for r in execute(query)
        )
        rows = []
        for row in results:
            for new_entry in self.generator(self.catalog, row):
                rows.append(new_entry)
                if len(rows) >= BATCH_SIZE:
                    execute(tbl.insert().values(rows))
                    rows.clear()
        if rows:
            execute(tbl.insert().values(rows))

        original_query = query.subquery()
        table_query = tbl.select().subquery()
        original_cols = [label(c.name, c) for c in original_query.columns]
        table_cols = [label(c.name, c) for c in table_query.columns]

        def q(*columns):
            names = {c.name for c in columns}
            cols1 = [c for c in original_cols if c.name in names]
            cols2 = [
                c for c in table_cols if c.name in names
            ]  # Columns for the generated table.
            q = sqlalchemy.union_all(
                sqlalchemy.select(*cols1).select_from(original_query),
                sqlalchemy.select(*cols2).select_from(table_query),
            )
            return q

        return QueryGenerator(q, [*original_cols])


class SQLFilter(Step):
    def __init__(self, *args):  # pylint: disable=super-init-not-called
        self.expressions = args

    def __and__(self, other):
        return self.__class__(*(self.expressions + other))

    def clone(self):
        return self.__class__(*self.expressions)

    def apply(self, query):
        new_query = query.filter(*self.expressions)

        def q(*columns):
            return new_query.with_only_columns(*columns)

        return QueryGenerator(q, new_query.selected_columns)


class SQLUnion(Step):
    def __init__(self, query1, query2):
        self.query1 = query1
        self.query2 = query2

    def apply(self, query):
        q1 = self.query1.apply_steps().select().subquery()
        q2 = self.query2.apply_steps().select().subquery()
        columns1, columns2 = fill_columns(q1.columns, q2.columns)

        def q(*columns):
            names = {c.name for c in columns}
            col1 = [c for c in columns1 if c.name in names]
            col2 = [c for c in columns2 if c.name in names]
            return (
                sqlalchemy.select(*col1)
                .select_from(q1)
                .union_all(sqlalchemy.select(*col2).select_from(q2))
            )

        return QueryGenerator(q, columns1)


def fill_columns(
    *column_iterables: Iterable["ColumnElement"],
) -> List[List["ColumnElement"]]:
    column_dicts = [{c.name: c for c in columns} for columns in column_iterables]
    combined_columns = {n: c for col_dict in column_dicts for n, c in col_dict.items()}

    result: List[List["ColumnElement"]] = [[] for _ in column_dicts]
    for n in combined_columns:
        col = next(col_dict[n] for col_dict in column_dicts if n in col_dict)
        for col_dict, out in zip(column_dicts, result):
            if n in col_dict:
                out.append(col_dict[n])
            else:
                # Cast the NULL to ensure all columns are aware of their type
                # Label it to ensure it's aware of its name
                out.append(sqlalchemy.cast(sqlalchemy.null(), col.type).label(n))
    return result


SQLQueryT = TypeVar("SQLQueryT", bound="SQLQuery")


class SQLQuery:
    def __init__(
        self,
        starting_step: StartingStep,
        steps: Optional[Iterable["Step"]] = None,
        catalog: Optional["Catalog"] = None,
        client_config=None,
    ):  # pylint: disable=super-init-not-called
        self.steps: List["Step"] = list(steps) if steps is not None else []
        self.starting_step: StartingStep = starting_step
        self.catalog = catalog or get_catalog(client_config=client_config)
        self._chunk_index: Optional[int] = None
        self._chunk_total: Optional[int] = None

    def __iter__(self):
        return iter(self.results())

    def __or__(self, other):
        return self.union(other)

    def apply_steps(self):
        """
        Apply the steps in the query and return the resulting
        sqlalchemy.Executable.
        """
        query = self.clone()

        index = os.getenv("DQL_QUERY_CHUNK_INDEX", self._chunk_index)
        total = os.getenv("DQL_QUERY_CHUNK_TOTAL", self._chunk_total)

        if None not in (index, total):
            index, total = int(index), int(total)  # os.getenv returns str

            if not (0 <= index < total):
                raise ValueError("chunk index must be between 0 and total")

            # Prepend the chunk filter to the step chain.
            query = query.filter(C.random % total == index)
            query.steps = query.steps[-1:] + query.steps[:-1]

        generator = query.starting_step.apply()
        for step in query.steps:
            generator = step.apply(
                generator.select()
            )  # a chain of steps linked by results
        return generator

    def results(self, row_factory=None):
        query = self.apply_steps()
        execute = self.catalog.data_storage.execute

        query = query.select()
        if row_factory:
            selected_columns = [col.name for col in query.selected_columns]
            result = [row_factory(selected_columns, r) for r in execute(query)]
        else:
            result = list(execute(query))
        return result

    def clone(self: SQLQueryT) -> SQLQueryT:
        obj = copy(self)
        obj.steps = obj.steps.copy()
        return obj

    def filter(self, *args):
        query = self.clone()
        steps = query.steps
        if steps and isinstance(steps[-1], SQLFilter):
            steps[-1] = steps[-1] & args
        else:
            steps.append(SQLFilter(*args))
        return query

    def union(self, dataset_query):
        left = self.clone()
        right = dataset_query.clone()
        new_query = self.clone()
        new_query.steps = [SQLUnion(left, right)]
        return new_query

    def chunk(self, index: int, total: int):
        """Split a query into smaller chunks for e.g. parallelization.
        Example:
            >>> query = DatasetQuery(...)
            >>> chunk_1 = query._chunk(0, 2)
            >>> chunk_2 = query._chunk(1, 2)
        Note:
            Bear in mind that `index` is 0-indexed but `total` isn't.
            Use 0/3, 1/3 and 2/3, not 1/3, 2/3 and 3/3.
        """
        query = self.clone()
        query._chunk_index, query._chunk_total = index, total
        return query


class DatasetQuery(SQLQuery):
    def __init__(
        self,
        path: str = "",
        name: str = "",
        version: Optional[int] = None,
        catalog=None,
        client_config=None,
    ):
        if catalog is None:
            catalog = get_catalog(client_config=client_config)

        data_storage = catalog.data_storage
        starting_step: StartingStep
        if path:
            starting_step = IndexingStep(path, catalog, client_config=client_config)
        elif name:
            ds = data_storage.get_dataset(name)
            version = version or ds.latest_version
            ds_table = catalog.data_storage.get_table(
                data_storage.dataset_table_name(dataset_id=ds.id, version=version)
            )
            starting_step = QueryStep(ds_table)
        else:
            raise ValueError("must provide path or name")

        super().__init__(
            starting_step=starting_step, catalog=catalog, client_config=client_config
        )

    def add_signal(self, udf: UDFType, name: Optional[str] = None):
        query = self.clone()
        steps = query.steps
        column_name = name or getattr(udf, "column", None)
        if not column_name:
            raise ValueError("column name not specified")
        steps.append(UDFSignal(column_name, udf, self.catalog))
        return query

    def generate(self, generator: "Generator"):
        query = self.clone()
        steps = query.steps
        steps.append(RowGenerator(generator, self.catalog))
        return query

    def save(self, name: str):
        """Save the query as a shadow dataset."""
        query = self.apply_steps()

        # Save to a temporary table first.
        temp_table_name = f"tmp_{name}_" + _random_string(6)
        columns: List["sqlalchemy.Column"] = [
            sqlalchemy.Column(col.name, col.type)
            for col in query.columns
            if col.name not in DATASET_CORE_COLUMN_NAMES
        ]
        self.catalog.data_storage.create_dataset_rows_table(
            temp_table_name,
            custom_columns=columns,
            if_not_exists=False,
        )
        temp_table = self.catalog.data_storage.get_table(temp_table_name)
        # Exclude the id column and let the db create it to avoid unique
        # constraint violations
        cols = [col.name for col in temp_table.c if col.name != "id"]

        self.catalog.data_storage.execute(
            sqlalchemy.insert(temp_table).from_select(cols, query.exclude("id"))
        )

        # Create a shadow dataset.
        self.catalog.data_storage.create_shadow_dataset(name, create_rows=False)
        dataset = self.catalog.data_storage.get_dataset(name)
        if dataset is None:
            raise RuntimeError(f"No dataset found with {name=}")
        # pylint: disable=protected-access
        table_name = self.catalog.data_storage.dataset_table_name(dataset.id)

        self.catalog.data_storage._rename_table(temp_table_name, table_name)
        self.catalog.data_storage.update_dataset_status(dataset, DatasetStatus.COMPLETE)


T = TypeVar("T")


def return_ds(dataset_query: T) -> T:
    if isinstance(dataset_query, DatasetQuery):
        ds_id = _random_string(6)
        ds_name = f"ds_return_{ds_id}"
        dataset_query.catalog.data_storage.return_ds_hook(ds_name)
        dataset_query.save(ds_name)
    return dataset_query


def _random_string(length: int) -> str:
    return "".join(
        random.choice(string.ascii_letters + string.digits)  # nosec B311
        for i in range(length)
    )



from datetime import datetime, timedelta, timezone

import pytest

from dql import utils
from dql.error import StorageNotFoundError
from dql.storage import STALE_HOURS_LIMIT, Status, Storage

TS = datetime(2022, 8, 1)
EXPIRES = datetime(2022, 8, 2)


def test_human_time():
    assert utils.human_time_to_int("1236") == 1236
    assert utils.human_time_to_int("3h") == 3 * 60 * 60
    assert utils.human_time_to_int("2w") == 2 * 7 * 24 * 60 * 60
    assert utils.human_time_to_int("4M") == 4 * 31 * 24 * 60 * 60

    assert utils.human_time_to_int("bla") is None


def test_storage():
    s = Storage("s3://foo", TS, EXPIRES)

    d = s.to_dict()
    assert d.get("uri") == s.uri


def test_expiration_time():
    assert Storage.get_expiration_time(TS, 12344) == TS + timedelta(seconds=12344)


def test_adding_storage(data_storage):
    # These tests re-use the same database for better performance
    data_storage.db.execute("DELETE FROM buckets")
    data_storage.db.commit()
    uri = "s3://whatever"
    with pytest.raises(StorageNotFoundError):
        data_storage.get_storage(uri)

    storage, _, _, _ = data_storage.register_storage_for_indexing(uri)
    cnt = data_storage.db.execute("SELECT COUNT() FROM buckets").fetchall()
    assert cnt[0][0] == 1

    query = "SELECT * FROM buckets WHERE uri = ?"
    bkt = data_storage.db.execute(query, [uri]).fetchall()
    assert len(bkt) == 1

    s = Storage(*bkt[0])
    assert s == storage


def test_storage_status(data_storage):
    uri = "s3://somebucket"

    data_storage.create_storage_if_not_registered(uri)
    storage = data_storage.get_storage(uri)
    assert storage.uri == uri
    assert storage.status == Status.CREATED

    (
        storage,
        need_index,
        in_progress,
        is_new,
    ) = data_storage.register_storage_for_indexing(uri)
    assert storage.status == Status.PENDING
    assert storage.uri == uri
    assert storage == data_storage.get_storage(uri)
    assert need_index is True
    assert in_progress is False
    assert is_new is True

    s2, need_index, in_progress, is_new = data_storage.register_storage_for_indexing(
        uri
    )
    assert s2.status == Status.PENDING
    assert storage == s2 == data_storage.get_storage(uri)
    assert need_index is False
    assert in_progress is True
    assert is_new is False

    end_time = datetime.now(timezone.utc)
    data_storage.mark_storage_indexed(uri, Status.COMPLETE, 1000, end_time)
    storage = data_storage.get_storage(uri)
    assert storage.status == Status.COMPLETE


@pytest.mark.parametrize(
    "ttl",
    (-1, 999999999999, 99999999999999, 9999999999999999),
)
def test_max_ttl(ttl):
    uri = "s3://whatever"
    expires = Storage.get_expiration_time(TS, ttl)
    storage = Storage(1, uri, TS, expires)
    assert storage.timestamp == TS
    assert storage.expires == datetime.max
    assert storage.timestamp_str  # no error
    assert storage.timestamp_to_local  # no error
    assert storage.expires_to_local  # no error


def test_storage_without_dates():
    uri = "s3://whatever"
    storage = Storage(1, uri, None, None)
    assert storage.timestamp is None
    assert storage.expires is None
    assert storage.timestamp_str is None  # no error
    assert storage.timestamp_to_local is None  # no error
    assert storage.expires_to_local is None  # no error
    assert storage.to_dict() == {
        "uri": uri,
        "timestamp": None,
        "expires": None,
    }


async def test_storage_update_last_inserted_at(data_storage):
    uri = "s3://bucket_last_inserted"
    data_storage.create_storage_if_not_registered(uri)
    await data_storage.update_last_inserted_at(uri)
    storage = data_storage.get_storage(uri)
    assert storage.last_inserted_at


def test_stale_storage(data_storage):
    uri_stale = "s3://bucket_stale"
    uri_not_stale = "s3://bucket_not_stale"

    data_storage.create_storage_if_not_registered(uri_stale)
    data_storage.create_storage_if_not_registered(uri_not_stale)

    data_storage.mark_storage_pending(data_storage.get_storage(uri_stale))
    data_storage.mark_storage_pending(data_storage.get_storage(uri_not_stale))

    # make storage looks stale
    updates = {
        "last_inserted_at": datetime.now(timezone.utc)
        - timedelta(hours=STALE_HOURS_LIMIT + 1)
    }
    s = data_storage.storages
    data_storage.execute(s.update().where(s.c.uri == uri_stale).values(**updates))

    data_storage.find_stale_storages()

    stale_storage = data_storage.get_storage(uri_stale)
    assert stale_storage.status == Status.STALE

    not_stale_storage = data_storage.get_storage(uri_not_stale)
    assert not_stale_storage.status == Status.PENDING



import os
import posixpath
from datetime import datetime, timezone
from pathlib import Path
from typing import Tuple

from fsspec.implementations.local import LocalFileSystem

from dql.data_storage import AbstractDataStorage
from dql.error import StorageNotFoundError
from dql.nodes_fetcher import NodesFetcher

from .checksum import file_digest
from .fsspec import FSSpecClient


class FileClient(FSSpecClient):
    FS_CLASS = LocalFileSystem
    PREFIX = "file://"
    protocol = "file"

    def __init__(
        self, name: str, fs: LocalFileSystem, use_symlinks: bool = False
    ) -> None:
        super().__init__(name, fs)
        self.use_symlinks = use_symlinks

    def url(self, path: str, expires: int = 3600) -> str:
        raise TypeError("Signed urls are not implemented for local file system")

    @classmethod
    def ls_buckets(cls, **kwargs):
        return []

    @classmethod
    def split_url(cls, url: str, data_storage) -> Tuple[str, str]:
        # pylint:disable=protected-access
        def _storage_exists(uri: str) -> bool:
            try:
                data_storage.get_storage(uri)
            except StorageNotFoundError:
                return False
            return True

        # lowercasing scheme just in case it's uppercase
        scheme, rest = url.split(":", 1)
        url = f"{scheme.lower()}:{rest}"
        if _storage_exists(url):
            return LocalFileSystem._strip_protocol(url), ""
        for pos in range(len(url) - 1, len(cls.PREFIX), -1):
            if url[pos] == "/" and _storage_exists(url[:pos]):
                return LocalFileSystem._strip_protocol(url[:pos]), url[pos + 1 :]
        raise RuntimeError(f"Invalid file path '{url}'")

    @classmethod
    def from_url(cls, url: str, data_storage, kwargs):
        result = cls(url, cls.create_fs(**kwargs))
        storage = data_storage.get_storage(result.uri)
        result.use_symlinks = storage.symlinks
        return result

    async def ls_dir(self, path):
        return self.fs.ls(path, detail=True)

    def rel_path(self, path):
        return posixpath.relpath(path, self.name)

    @property
    def uri(self):
        return Path(self.name).as_uri()

    def get_full_path(self, rel_path):
        full_path = Path(self.name, rel_path).as_uri()
        if rel_path.endswith("/") or not rel_path:
            full_path += "/"
        return full_path

    def _dict_from_info(self, v, parent_id, path, partial_id):
        name = posixpath.basename(path)
        return {
            "dir_id": None,
            "parent_id": parent_id,
            "path": self.rel_path(v["name"]),
            "name": name,
            "checksum": "",
            "etag": "",
            "version": "",
            "is_latest": True,
            "last_modified": datetime.fromtimestamp(v["mtime"], timezone.utc),
            "size": v.get("size", ""),
            "owner_name": "",
            "owner_id": "",
            "anno": None,
            "partial_id": partial_id,
        }

    def fetch_nodes(
        self,
        file_path,
        nodes,
        cache,
        data_storage: AbstractDataStorage,
        total_size=None,
        cls=NodesFetcher,
        pb_descr="Download",
        shared_progress_bar=None,
    ):
        if self.use_symlinks:
            return self._fetch_symlinks(nodes, data_storage)
        else:
            return super().fetch_nodes(
                file_path,
                nodes,
                cache,
                data_storage,
                total_size,
                cls,
                pb_descr,
                shared_progress_bar,
            )

    def _fetch_symlinks(self, nodes, data_storage):
        updated_nodes = []
        for node in nodes:
            if not node.is_downloadable:
                continue
            src_path = f"{self.name}/{node.path_str}"
            with open(src_path, "rb") as fileobj:
                md5_hash = file_digest(fileobj)
            data_storage.update_checksum(node, md5_hash)
            node.checksum = md5_hash
            updated_nodes.append(node)
        return updated_nodes

    def do_instantiate_node(self, node, cache, dst):
        if self.use_symlinks:
            os.symlink(Path(self.name, node.path_str), dst)
        else:
            super().do_instantiate_node(node, cache, dst)



import json
import os
import tarfile
import tempfile
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from itertools import groupby, islice
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Tuple, Union

from sqlalchemy import Float, Integer, String
from tqdm import tqdm

from dql.node import DirType

if TYPE_CHECKING:
    from dql.listing import Listing


PROCESSING_BATCH_SIZE = 1000  # Batch size for inserting entries.

INSERT_ITEMS = "insert"  # Indexing format adds new objects.
UPDATE_ITEMS = "update"  # Indexing format extends objects with new properties.


class IndexingFormat(ABC):
    """
    Indexing formats allow additional transformations on indexed
    objects, such as listing contents of archives.
    """

    # Default processor action - add new items to the index or update them,
    # e.g. by adding new signals.
    action_type: str = INSERT_ITEMS

    @abstractmethod
    def filter(self, listing: "Listing", paths: List[str]) -> Iterator[Tuple[Any, ...]]:
        """Create a list of entries to process"""

    @abstractmethod
    def process(self, listing, entries):
        """Process an entry and return additional entries to store."""


class TarFiles(IndexingFormat):
    """
    TarFiles indexes buckets containing uncompressed tar archives. The contents of
    the archives is indexed as well.
    """

    item_type = INSERT_ITEMS

    def filter(self, listing: "Listing", paths: List[str]) -> Iterator[Tuple[Any, ...]]:
        for path in paths:
            for node in listing.expand_path(path):
                found = listing.find(
                    node,
                    ["path_str", "id", "is_latest", "partial_id"],
                    names=["*.tar"],
                )
                yield from found

    def process(self, listing: "Listing", entries):
        for entry in entries:
            yield from self.process_entry(listing, entry)

    def process_entry(self, listing: "Listing", entry):
        pth, parent_id, is_latest, partial_id = entry
        local_path = tempfile.gettempdir() + f"/dql_cache_{parent_id}"
        client = listing.client
        # Download tarball to local storage first.
        client.fs.get_file(client.get_full_path(pth), local_path)
        with tarfile.open(name=local_path, mode="r:") as tar:
            for info in tar:
                if info.isdir():
                    yield self.tardir_from_info(
                        info, parent_id, pth, is_latest, partial_id
                    )
                elif info.isfile():
                    yield self.tarmember_from_info(
                        info, parent_id, pth, is_latest, partial_id
                    )
        os.remove(local_path)
        listing.data_storage.update_type(parent_id, DirType.TAR_ARCHIVE)

    def tarmember_from_info(self, info, parent_id, path_str, is_latest, partial_id):
        location = json.dumps(
            [
                {
                    "offset": info.offset_data,
                    "size": info.size,
                    "type": "tar",
                    "parent": path_str,
                },
            ]
        )
        return {
            "vtype": "tar",
            "dir_type": DirType.FILE,
            "parent_id": parent_id,
            "path_str": f"{path_str}/{info.name}",
            "name": info.name.rsplit("/", 1)[-1],
            "checksum": "",
            "etag": "",
            "version": "",
            "is_latest": is_latest,
            "last_modified": datetime.fromtimestamp(info.mtime, timezone.utc),
            "size": info.size,
            "owner_name": info.uname,
            "owner_id": str(info.uid),
            "location": location,
            "partial_id": partial_id,
        }

    def tardir_from_info(self, info, parent_id, path_str, is_latest, partial_id):
        return {
            "vtype": "tar",
            "dir_type": DirType.DIR,
            "parent_id": parent_id,
            "path_str": f"{path_str}/{info.name.rstrip('/')}",
            "name": info.name.rsplit("/", 1)[-1],
            "checksum": "",
            "etag": "",
            "version": "",
            "is_latest": is_latest,
            "last_modified": datetime.fromtimestamp(info.mtime, timezone.utc),
            "size": info.size,
            "owner_name": info.uname,
            "owner_id": str(info.uid),
            "partial_id": partial_id,
        }


class LaionJSONPair(IndexingFormat):
    """
    Load signals from .json files and attach them to objects with the same base name.
    """

    action_type = UPDATE_ITEMS

    IGNORED_EXTS = [".json", ".txt"]  # File extensions not to attach loaded signals to.

    def filter(self, listing: "Listing", paths: List[str]) -> Iterator[Tuple[Any, ...]]:
        for path in paths:
            for node in listing.expand_path(path):
                found = listing.find(
                    node,
                    ["path_str", "id", "vtype", "name", "size", "location"],
                    order_by="path_str",
                )
                yield from found

    def process(self, listing: "Listing", entries):
        self._create_columns(listing)
        for _, group in groupby(entries, self._group):
            yield from self._process_group(listing, group)

    def _process_group(self, listing: "Listing", group):
        # Create a map of extension to object info.
        nodes = {
            os.path.splitext(entry[0])[1]: {
                "path_str": entry[0],
                "id": entry[1],
                "vtype": entry[2],
                "name": entry[3],
                "size": entry[4],
                "location": entry[5],
            }
            for entry in group
        }
        json_obj = nodes.get(".json")
        if not json_obj:
            # No .json file in group. Ignore.
            return
        with listing.client.open_object(
            json_obj["path_str"],
            json_obj["vtype"],
            **{arg: json_obj[arg] for arg in ("name", "size", "location")},
        ) as f:
            data = json.load(f)
            if not isinstance(data, dict):
                # .json file contains something other than a json object. Ignore.
                return
            update = {
                col[0]: data.get(col[0]) for col in laionJSONColumns if data.get(col[0])
            }

        for ext, node in nodes.items():
            if ext in self.IGNORED_EXTS:
                continue
            yield (node["id"], update)

    def _group(self, entry):
        """
        Group entries by paths sans the extension.
        This way 'path/000.jpg' and 'path/000.json' will be grouped
        together.
        """
        return os.path.splitext(entry[0])[0]

    def _create_columns(self, listing: "Listing"):
        for col_name, col_type in laionJSONColumns:
            listing.data_storage.add_bucket_signal_column(
                listing.client.uri, col_name, col_type
            )


# Signal columns for storing laion json data.
laionJSONColumns = (
    ("punsafe", Float()),
    ("pwatermark", Float()),
    ("similarity", Float()),
    ("hash", Integer()),
    ("caption", String()),
    ("url", String()),
    ("key", String()),
    ("status", String()),
    ("error_message", String()),
    ("width", Integer()),
    ("height", Integer()),
    ("original_width", Integer()),
    ("original_height", Integer()),
    ("md5", String()),
)


async def apply_processors(
    listing: "Listing", path: str, processors: List[IndexingFormat]
):
    async def insert_items(items):
        await listing.data_storage.insert_entries(items)

    async def update_items(items):
        for item in items:
            (node_id, values) = item
            listing.data_storage.update_node(node_id, values)

    for processor in processors:
        if processor.action_type == INSERT_ITEMS:
            func = insert_items
        elif processor.action_type == UPDATE_ITEMS:
            func = update_items

        listing = listing.clone()
        with tqdm(desc="Processing", unit=" objects") as pbar:
            entries = processor.filter(listing, [path])
            results = processor.process(listing, entries)
            for batch in _batch(results, PROCESSING_BATCH_SIZE):
                pbar.update(len(batch))
                await func(batch)
        if processor.action_type == INSERT_ITEMS:
            listing.data_storage.inserts_done()


def _batch(it, size):
    while batch := list(islice(it, size)):
        yield batch


indexer_formats: Dict[str, Union[List[IndexingFormat], IndexingFormat]] = {
    "tar-files": TarFiles(),
    "json-pair": LaionJSONPair(),
    "webdataset": [TarFiles(), LaionJSONPair()],
}



from sqlalchemy import select
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import true

from ..data import ENTRIES, INVALID_ENTRIES
from ..utils import insert_entries


def test_validate_paths(data_storage):
    src = "s3://bucket1"
    data_storage = data_storage.clone(uri=src)
    data_storage.init_db(src, True)
    insert_entries(data_storage, ENTRIES + INVALID_ENTRIES)

    def count_valid_nodes():
        n = data_storage.nodes
        query = select(func.count(n.c.valid)).where(  # pylint: disable=not-callable
            n.c.valid == true()
        )
        return data_storage.execute(query).fetchone()[0]

    before_count = count_valid_nodes()
    data_storage.validate_paths()
    after_count = count_valid_nodes()

    assert before_count == 14
    assert after_count == 11



from typing import TYPE_CHECKING, Any, Callable, Type, Union

from dql.catalog import Catalog

from .schema import Column, Object

if TYPE_CHECKING:
    from dql.dataset import DatasetRow

UDFType = Callable[["Catalog", "DatasetRow"], Any]


def udf(output_type: Any, *parameters: Union["Column", "Object"]):
    """Decorate a function to be usable as a UDF."""

    def decorator(func: Callable):
        return UDF(func, output_type, *parameters)

    return decorator


class UDF:
    """A wrapper class for UDFs to be used in custom signal generation."""

    def __init__(
        self,
        func: Callable,
        output_type: Any,
        *parameters: Union["Column", "Object"],
    ):
        self.func = func
        self.output_type = output_type
        self.parameters = parameters

    def __call__(self, catalog: "Catalog", row: "DatasetRow") -> Any:
        params = []
        for p in self.parameters:
            if isinstance(p, Column):
                params.append(row[p.name])
            elif isinstance(p, Object):
                with catalog.open_object(row) as f:
                    obj: Any = p.reader(f)
                params.append(obj)
            else:
                raise ValueError("unknown udf parameter")
        return self.func(*params)


def generator(*parameters: Union["Column", "Object", Type["Catalog"]]):
    def decorator(func: Callable):
        return Generator(func, *parameters)

    return decorator


class Generator:
    """A wrapper class for UDFs used to generate new dataset rows."""

    def __init__(
        self, func: Callable, *parameters: Union["Column", "Object", Type["Catalog"]]
    ):
        self.func = func
        self.parameters = parameters

    def __call__(self, catalog: "Catalog", row: "DatasetRow"):
        params = []
        for p in self.parameters:
            if isinstance(p, Column):
                params.append(row[p.name])
            elif isinstance(p, Object):
                with catalog.open_object(row) as f:
                    obj: Any = p.reader(f)
                params.append(obj)
            elif p is Catalog:
                params.append(catalog)
            else:
                raise ValueError("unknown udf parameter")
        yield from self.func(row, *params)
